from .variables import Variable
import copy


class VariablePool(list):
    """
    Class for list of variables. All variables should have different names.

    :param lst: list of the variables.
    """
    def __init__(self, lst=None):
        self.names = set()
        if lst is not None:
            for item in lst:
                self.append(item)

    def fix_pickling(self):
        if not hasattr(self, 'names'):
            self.names = set()
            for item in self:
                self.names.append(item.name)

    def check_type(self, item):
        """
        Raises ValueError if ``item`` is not Variable.
        """
        if not isinstance(item, Variable):
            raise ValueError(f"Items of VariablePool could be Variables only.")

    def append(self, item):
        """
        Append new ``item`` to pool.
        """
        self.fix_pickling()
        self.check_type(item)
        if item.name not in self.names:
            self.names.add(item.name)
            super(VariablePool, self).append(item)
        else:
            raise NameError(f"VariablePool has already a Variable with "
                            f"the same name ({item.name}).")

    def extend(self, items):
        """
        Extend pool with other list.
        """
        for item in items:
            self.append(item)

    def __setitem__(self, key, item):
        self.fix_pickling()
        if isinstance(key, slice):
            remove_names = []
            for value in self[key]:
                remove_names.append(value.name)
            for value in item:
                self.check_type(value)
                if value.name in self.names and value.name not in remove_names:
                    raise NameError(f"VariablePool has already a Variable"
                                    f" with the same name ({key}).")
            for name in remove_names:
                self.names.remove(name)
            for value in item:
                self.names.add(value.name)
            return super(VariablePool, self).__setitem__(key, item)

        self.check_type(item)
        if item.name not in self.names or item.name == self[key].name:
            self.names.remove(self[key].name)
            self.names.add(item.name)
            return super(VariablePool, self).__setitem__(key, item)
        else:
            raise NameError(f"VariablePool has already a Variable with "
                            f"the same name ({key}).")

    def __delitem__(self, key):
        self.fix_pickling()
        if isinstance(key, slice):
            for value in self[key]:
                self.names.remove(value.name)
        else:
            value = self[key]
            self.names.remove(value.name)
        super(VariablePool, self).__delitem__(key)

    def __copy__(self):
        newone = type(self)(self)
        return newone

    def __deepcopy__(self, memo):
        lst = list()
        for var in self:
            lst.append(copy.deepcopy(var, memo))
        newone = type(self)(lst)
        return newone



import numpy as np


def trunc_normal(mean, sigma, lower, upper):
    """
    Truncated normal distribution.
    """
    from scipy.stats import truncnorm
    if sigma == 0:
        sigma = 1e-15
    a, b = (lower - mean) / sigma, (upper - mean) / sigma
    return truncnorm.rvs(a, b, loc=mean, scale=sigma)


def trunc_lognormal(mean, sigma, lower, upper):
    """
    Truncated log-normal distribution.
    """
    mean_plus_sigma = mean + sigma
    transf_mean = np.log(mean)
    transf_sigma = np.log(mean_plus_sigma) - transf_mean
    return np.exp(trunc_normal(transf_sigma, transf_sigma,
                               np.log(lower), np.log(upper)))


def trunc_normal_3_sigma_rule(mean, lower, upper):
    """
    Truncated normal distribution with sigma according by
    three sigma rule.
    """
    sigma = min(mean - lower, upper - mean) / 3
    return trunc_normal(mean, sigma, lower, upper)


def trunc_lognormal_3_sigma_rule(mean, lower, upper):
    """
    Truncated log-normal distribution with sigma according by
    three sigma rule.
    """
    return np.exp(trunc_normal_3_sigma_rule(np.log(mean), np.log(lower),
                                            np.log(upper)))


# Random generators
def uniform_generator(domain):
    """
    Uniform generator. Runs ``numpy.random.uniform`` on ``domain``.
    """
    return np.random.uniform(domain[0], domain[1])


def trunc_lognormal_sigma_generator(domain):
    """
    Generator for :func:`trunc_lognormal_3_sigma_rule`
    """
    lower = domain[0]
    if domain[0] == 0:
        lower = 1e-15
    mean = 1
    if mean < domain[0]:
        mean = domain[0]
    if mean > domain[1]:
        mean = domain[1]
    return trunc_lognormal_3_sigma_rule(mean, lower, domain[1])


def trunc_normal_sigma_generator(domain):
    """
    Generator for :func:`trunc_normal_3_sigma_rule`
    """
    mean = 1
    if mean < domain[0]:
        mean = domain[0]
    if mean > domain[1]:
        mean = domain[1]
    return trunc_normal_3_sigma_rule(mean, domain[0], domain[1])


def custom_generator(variables):
    """
    Custom generator for demographic model variables.
    """
    from .variables import PopulationSizeVariable, ContinuousVariable
    N_A = PopulationSizeVariable("__name").resample()
    values = list()
    for var in variables:
        x = var.resample()
        values.append(var.translate_units(x, N_A))
        if isinstance(var, ContinuousVariable):
            values[-1] = max(values[-1], var.domain[0])
            values[-1] = min(values[-1], var.domain[1])
    return np.array(values, dtype=object)


# def multiply_generator(gen1, domain1, gen2, domain2):
#     def generator(domain):
#         return gen1(domain1) * gen2(domain2)
#     return generator



from . import Engine, register_engine, get_engine
from .dadi_moments_common import DadiOrMomentsEngine
from ..models import DemographicModel, CustomDemographicModel, Epoch, Split
from ..utils import DynamicVariable
from .. import SFSDataHolder


class DadiEngine(DadiOrMomentsEngine):
    """
    Engine for using :py:mod:`dadi` for demographic inference.

    Citation of :py:mod:`dadi`:

    Gutenkunst RN, Hernandez RD, Williamson SH, Bustamante CD (2009)
    Inferring the Joint Demographic History of Multiple Populations
    from Multidimensional SNP Frequency Data. PLoS Genet 5(10): e1000695.
    https://doi.org/10.1371/journal.pgen.1000695
    """

    id = 'dadi'  #:
    import dadi as base_module
    supported_data = [SFSDataHolder]  #:
    inner_data_type = base_module.Spectrum  #:

    @staticmethod
    def _get_kwargs(event, var2value):
        """
        Builds kwargs for dadi.Integration functions (one_pop, two_pops,
        three_pops).

        :param event: build for this event
        :type event: event.Epoch
        :param var2value: dictionary {variable: value}, it is required because
            the dynamics values should be fixed.
        """
        ret_dict = {'T': event.time_arg}
        for i in range(event.n_pop):
            if event.n_pop == 1:
                arg_name = 'nu'
            else:
                arg_name = 'nu%d' % (i+1)

            if event.dyn_args is not None:
                dyn_arg = event.dyn_args[i]
                dyn = DadiEngine.get_value_from_var2value(var2value, dyn_arg)
                if dyn == 'Sud':
                    ret_dict[arg_name] = event.size_args[i]
                else:
                    ret_dict[arg_name] = 'nu%d_func' % (i+1)
            else:
                ret_dict[arg_name] = event.size_args[i]

        if event.mig_args is not None:
            for i in range(event.n_pop):
                for j in range(event.n_pop):
                    if i == j:
                        continue
                    ret_dict['m%d%d' % (i+1, j+1)] = event.mig_args[i][j]
        if event.sel_args is not None:
            if event.n_pop == 1:
                arg_name = 'gamma'
            else:
                arg_name = 'gamma%d' % (i+1)
            for i in range(event.n_pop):
                ret_dict[arg_name] = event.sel_args[i]
        if event.dom_args is not None:
            if event.n_pop == 1:
                arg_name = 'h'
            else:
                arg_name = 'h%d' % (i+1)
            for i in range(event.n_pop):
                ret_dict[arg_name] = event.dom_args[i]
        return ret_dict

    def _inner_func(self, values, ns, pts):
        """
        Simulates expected SFS for proposed values of variables.

        :param values: values of variables
        :param ns: sample sizes of simulated SFS
        :param pts: grid points for numerical solution
        """
        if isinstance(self.model, CustomDemographicModel):
            return self.model.function(values, ns, pts)

        var2value = self.model.var2value(values)
        dadi = self.base_module

        xx = dadi.Numerics.default_grid(pts)
        phi = dadi.PhiManip.phi_1D(xx)

        addit_values = {}
        for ind, event in enumerate(self.model.events):
            if isinstance(event, Epoch):
                if event.dyn_args is not None:
                    for i in range(event.n_pop):
                        dyn_arg = event.dyn_args[i]
                        value = self.get_value_from_var2value(var2value,
                                                              dyn_arg)
                        if value != 'Sud':
                            func = DynamicVariable.get_func_from_value(value)
                            y1 = self.get_value_from_var2value(
                                var2value, event.init_size_args[i])
                            y2 = self.get_value_from_var2value(
                                var2value, event.size_args[i])
                            x_diff = self.get_value_from_var2value(
                                var2value, event.time_arg)
                            addit_values['nu%d_func' % (i+1)] = func(
                                y1=y1,
                                y2=y2,
                                x_diff=x_diff)
                kwargs_with_vars = self._get_kwargs(event, var2value)
                kwargs = {x: self.get_value_from_var2value(var2value, y)
                          for x, y in kwargs_with_vars.items()}
                kwargs = {x: addit_values.get(y, y)
                          for x, y in kwargs.items()}
                if event.n_pop == 1:
                    phi = dadi.Integration.one_pop(phi, xx, **kwargs)
                if event.n_pop == 2:
                    phi = dadi.Integration.two_pops(phi, xx, **kwargs)
                if event.n_pop == 3:
                    phi = dadi.Integration.three_pops(phi, xx, **kwargs)
            elif isinstance(event, Split):
                if event.n_pop == 1:
                    phi = dadi.PhiManip.phi_1D_to_2D(xx, phi)
                else:
                    func_name = "phi_%dD_to_%dD_split_%d" % (
                        event.n_pop, event.n_pop + 1, event.pop_to_div + 1)
                    phi = getattr(dadi.PhiManip, func_name)(xx, phi)
        sfs = dadi.Spectrum.from_phi(phi, ns, [xx]*len(ns))
        return sfs

    def draw_schematic_model_plot(self, values, save_file=None,
                                  fig_title="Demographic Model from GADMA",
                                  nref=None, gen_time=1,
                                  gen_time_units="Generations"):
        """
        Draws schematic plot of the model with values with
        :class:`MomentsEngine` (!).
        See moments manual for more information.

        :param values: Values of the model parameters, it could be list of
                       values or dictionary {variable name: value}.
        :type values: list or dict
        :param save_file: File to save picture. If None then picture will be
                          displayed to the screen.
        :type save_file: str
        :param fig_title: Title of the figure.
        :type fig_title: str
        :param nref: An ancestral population size. If None then parameters
                     will be drawn in genetic units.
        :type nref: int
        :param gen_type: Time of one generation.
        :type gen_type: float
        :param gen_time_units: Units of `gen_type`. For example, it
                               could be Years, Generations, Thousand Years and
                               so on.
        """
        if isinstance(self.model, CustomDemographicModel):
            raise RuntimeError("Could not draw model plot for dadi's "
                               "custom file.")
        try:
            moments_engine = get_engine('moments')
        except AttributeError:
            raise RuntimeError("Moments Engine is requiered to draw schematic"
                               " model plot.")
        moments_engine.set_data(self.data)
        moments_engine.set_model(self.model)
        moments_engine.draw_schematic_model_plot(values, save_file, fig_title,
                                                 nref, gen_time,
                                                 gen_time_units)

    def simulate(self, values, ns, pts):
        """
        Returns simulated expected SFS for :attr:`demographic_model` with
        values as parameters. Simulation is performed with :attr:`self.pts`
        as grid points for numerical solutions.

        :param values: values of demographic model parameters.
        :param ns: sample sizes of the simulated SFS.
        """
        dadi = self.base_module
        func_ex = dadi.Numerics.make_extrap_log_func(self._inner_func)
#        print(values, ns, pts)
        model = func_ex(values, ns, pts)
        # TODO: Nref
        return model

    def get_theta(self, values, pts):
        return super(DadiEngine, self).get_theta(values, pts)

    def evaluate(self, values, pts):
        return super(DadiEngine, self).evaluate(values, pts)

    def generate_code(self, values, filename, pts,
                      nanc=None, gen_time=None, gen_time_units=None):
        return super(DadiEngine, self).generate_code(values, filename, pts,
                                                     nanc, gen_time,
                                                     gen_time_units)


register_engine(DadiEngine)



from . import Engine
from ..models import DemographicModel, StructureDemographicModel,\
                     CustomDemographicModel, Epoch, Split
from ..utils import DynamicVariable, DiscreteVariable, cache_func
from .. import SFSDataHolder
from .. import dadi_available, moments_available
from ..code_generator import id2printfunc

import warnings
import os
import numpy as np
from functools import wraps


class DadiOrMomentsEngine(Engine):
    """
    Base engine class for dadi and moments engines. It has all common methods
    of both engines.
    """

    id = 'dadi_or_moments'  #:
    base_module = None  #:
    supported_models = [DemographicModel, StructureDemographicModel,
                        CustomDemographicModel]  #:
    supported_data = [SFSDataHolder]  #:
    inner_data_type = None  # base_module.Spectrum  #:

    @classmethod
    def read_data(cls, data_holder):
        """
        Reads SFS data from `data_holder`.

        Could read two types of data:

            * :py:mod:`dadi` SFS data type
            * :py:mod:`dadi` SNP data type

        Check :py:mod:`dadi` manual for additional information.

        :param data_holder: holder of the data.
        :type data_holder: :class:`SFSDataHolder`
        """
        if data_holder.__class__ not in cls.supported_data:
            raise ValueError(f"Data class {data_holder.__class__.__name__}"
                             f" is not supported by {cls.id} engine.\nThe "
                             f"supported classes are: {cls.supported_data}"
                             f" and {cls.inner_data_type}")
        data = read_dadi_data(cls.base_module, data_holder)
        return data

    def _get_key(self, x, grid_sizes):
        var2value = self.model.var2value(x)
        key = tuple(var2value[var] for var in self.model.variables)
        if isinstance(grid_sizes, float):
            return (key, grid_sizes)
        return (key, tuple(grid_sizes))

    def _get_theta_from_sfs(self, values, model_sfs):
        optimal_theta = self.base_module.Inference.optimal_sfs_scaling(
            model_sfs, self.data)
        Nanc = self.get_N_ancestral_from_theta(1.0) or 1.0
        theta0 = 1 / Nanc
        if self.model.linear_constrain is None:
            return optimal_theta
        # If we have constrains we deal with them:
        upper_lb = None
        lower_ub = None
        Ax = self.model.linear_constrain._get_value(values)
        theta_lb = theta0 * self.model.linear_constrain.lb / Ax
        theta_ub = theta0 * self.model.linear_constrain.ub / Ax
        if np.any(theta_lb > theta_ub):
            inds = np.where(theta_lb > theta_ub)
            raise ValueError(f"Lower bounds for {inds} constrains in model"
                             f" are greater than upper bounds."
                             f"Please check linear constrain:\n"
                             f"{self.model.linear_constrain}\n"
                             f"and values:\n{values}.")
        theta_upper_lb = max(theta_lb)
        theta_lower_ub = min(theta_ub)
        if theta_upper_lb > theta_lower_ub:
            raise ValueError(f"Upper lower bound ({upper_lb}) in greater than"
                             f" lower upper bound ({lower_ub}). Please check "
                             f" linear constrain:\n"
                             f"{self.model.linear_constrain}\n"
                             f"and values:\n{values}.")

        optimal_theta = max(optimal_theta, theta_upper_lb)
        optimal_theta = min(optimal_theta, theta_lower_ub)
        return optimal_theta
#        for i, (lb, ub, val) in enumerate(zip(self.model.linear_constrain.lb,
#                                              self.model.linear_constrain.ub,
#                                              Ax):
#            if lb > ub:
#                raise ValueError(f"Lower bound for {i} constrain in model for"
#                                 f" values {values} is greater than upper "
#                                 f"bound: {val}, [{low_bound}, {upp_bound}]."
#                                 f"Please check linear constrain:\n"
#                                 f"{self.model.linear_constrain}\n"
#                                 f"and values:\n{values}.")
#            if lb is not None:
#                if upper_lb is None:
#                    upper_lb = lb / val
#                upper_lb = max(upper_lb, lb / val)
#            if ub is not None:
#                if lower_ub is None:
#                    lower_ub = ub / val
#                lower_ub = max(lower_ub, ub / val)
#        if upper_lb > lower_ub:
#            raise ValueError(f"Upper lower bound ({upper_lb}) in greater than"
#                             f" lower upper bound ({lower_ub}). Please check "
#                             f" linear constrain:\n"
#                             f"{self.model.linear_constrain}\n"
#                             f"and values:\n{values}.")
#        optimal_theta = max(optimal_theta, lower_ub)

    def get_theta(self, values, grid_sizes):
        key = self._get_key(values, grid_sizes)
        if key not in self.saved_add_info:
            warnings.warn("Additional evaluation for theta. Nothing to worry "
                          "if this warning is seldom.")
            self.evaluate(values, grid_sizes)
        theta = self.saved_add_info[key]
        return theta

    def get_N_ancestral_from_theta(self, theta):
        if self.model.theta0 is not None:
            theta0 = self.model.theta0
        elif (self.model.mu is not None and self.data_holder is not None and
                self.data_holder.sequence_length is not None):
            theta0 = 4 * self.model.mu * self.data_holder.sequence_length
        else:
            return None
        return theta / theta0

    def get_N_ancestral(self, values, grid_sizes):
        theta = self.get_theta(values, grid_sizes)
        return self.get_N_ancestral_from_theta(theta)

    def draw_sfs_plots(self, values, grid_sizes, save_file=None, vmin=None):
        """
        Draws plots of SFS data for observed data and simulated by model data.

        :param values: Values of the model parameters, it could be list of
                       values or dictionary {variable name: value}.
        :type values: list or dict
        :param grid_sizes: special parameter for numerical solutions. It is
                           `pts` for :class:`DadiEngine` and
                           `dt_fac` for :class:`MomentsEngine`.
        :param save_file: File to save picture. If None then picture will be
                          displayed to the screen.
        :type save_file: str
        :param vmin: Values in sfs below vmin are masked in plot. Should be
                     positive.
        :type vmin: float
        """
        from matplotlib import pyplot as plt
        plt.clf()
        plt.close()
        if vmin is not None:
            assert vmin > 0
        show = save_file is None
        n_pop = len(self.data.sample_sizes)
        # Get simulated data
        model = self.simulate(values, self.data.sample_sizes, grid_sizes)
        # Draw
        if n_pop == 1:
            self.base_module.Plotting.plot_1d_comp_multinom(model, self.data)
            if show:
                plt.show()
        else:
            func_name = f"plot_{n_pop}d_comp_multinom"
            plotting_func = getattr(self.base_module.Plotting, func_name)
            plotting_func(model, self.data, vmin=vmin, show=show)
        if not show:
            plt.savefig(save_file)
            plt.close('all')

    def evaluate(self, values, grid_sizes):
        """
        Simulates SFS from values and evaluate log likelihood between
        simulated SFS and observed SFS.
        """
        if self.data is None or self.model is None:
            raise ValueError("Please set data and model for the engine or"
                             " use set_and_evaluate function instead.")
        model_sfs = self.simulate(values, self.data.sample_sizes, grid_sizes)
        # The next two lines usually works like ll_multinom, but when we have
        # some constrains it could turn out to be ll with some other theta.
        theta = self._get_theta_from_sfs(values, model_sfs)
        ll_model = self.base_module.Inference.ll(theta * model_sfs, self.data)
        # Save simulated data
        key = self._get_key(values, grid_sizes)
        self.saved_add_info[key] = theta
        return ll_model

    def get_claic_component(self, x0, all_boots, grid_sizes, eps):
        if dadi_available:
            from dadi import Godambe
        elif moments_available:
            from moments import Godambe
        else:
            ImportError("For CLAIC evalueation either dadi or moments is"
                        " required.")
        # Cache evaluations of the frequency spectrum inside our hessian/J
        # evaluation function
        var2val = self.model.var2value(x0)
        is_not_discrete = np.array([not isinstance(var, DiscreteVariable)
                                    for var in var2val])
        if len(x0) > 0 and len(var2val) > 0:
            x0 = np.array(list(var2val.values()), dtype=object)
            p0 = x0[is_not_discrete]
        else:
            p0 = x0

        @wraps(self.simulate)
        def simul_func(x):
            p = np.array(x0)
            if len(p) > 0 and len(var2val) > 0:
                p[is_not_discrete] = x
            else:
                p = x
            return self.simulate(p, self.data.sample_sizes, grid_sizes)

        cached_simul = cache_func(simul_func)

        def func(x, data):
            model = cached_simul(x)
            return self.base_module.Inference.ll_multinom(model, self.data)
        H = - Godambe.get_hess(func, p0, eps, args=[self.data])
        H_inv = np.linalg.inv(H)

        J = np.zeros((len(p0), len(p0)))
        for ii, boot in enumerate(all_boots):
            boot = self.base_module.Spectrum(boot)
            grad_temp = Godambe.get_grad(func, p0, eps, args=[boot])
            J_temp = np.outer(grad_temp, grad_temp)
            J += J_temp

        J = J/len(all_boots)

        # G = J*H^-1
        G = np.dot(J, H_inv)

        return np.trace(G)

    def generate_code(self, values, filename, grid_sizes, nanc=None,
                      gen_time=None, gen_time_units="years"):
        """
        Prints nice formated code in the format of engine to file or returns
        it as string if no file is set.

        :param values: values for the engine's model.
        :param filename: file to print code. If None then the string will
                         be returned.
        """
        if self.data_holder is None:
            raise AttributeError("Engine was initialized with inner "
                                 "data. Need gadma.DataHolder for "
                                 "generation of code.")
        return id2printfunc[self.id](self, values,
                                     grid_sizes, filename, nanc, gen_time,
                                     gen_time_units)


# Those functions are common for dadi and moments engines.
def _check_missing_population_labels(sfs, default_pop_labels=None,
                                     filename=None):
    """
    Check that SFS has population labels. If not then make them default.

    : param sfs: site frequency spectrum to check
    : type sfs: dadi.Spectrum (or analogue)
    : param default_population_labels: if pop. labels are missing use
                                       this values.
    """
    if sfs.pop_ids is None:
        if default_pop_labels is not None:
            warnings.warn("Spectrum file %s is in an old format - without"
                          " population labels, so they will be taken from"
                          " corresponding parameter: %s."
                          % (filename, ', '.join(default_pop_labels)))
            sfs.pop_ids = default_pop_labels
        else:
            sfs.pop_ids = ['Pop %d' % (i+1) for i in range(sfs.ndim)]
    return sfs


def _new_population_labels(sfs, new_labels):
    """
    Assign new order of population labels of SFS.

    : param sfs: site frequency spectrum to change its labels
    : type sfs: dadi.Spectrum (or analogue)
    : param new_labels: new population labels
    """
    if new_labels is None:
        return sfs
    if sfs.pop_ids != new_labels:
        # Create a permutation of axis
        d = {x: i for i, x in enumerate(sfs.pop_ids)}
        try:
            d = [d[x] for x in new_labels]
        except:  # NOQA
            raise ValueError("Wrong Population labels parameter, population"
                             " labels are: " + ', '.join(sfs.pop_ids))
        # Rotate axis
        sfs = np.transpose(sfs, d)
        sfs.pop_ids = new_labels
    return sfs


def _project(sfs, new_size):
    """
    Project SFS to new sample size.

    : param sfs: site frequency spectrum to change its size
    : type sfs: dadi.Spectrum (or analogue)
    : param new_size: new sample size
    : type new_size: np.ndarray
    """
    if new_size is None:
        return sfs
    if not list(new_size) == list(sfs.sample_sizes):
        try:
            sfs = sfs.project(np.array(new_size))
        except Exception as e:
            raise ValueError("Wrong projections of SFS: " + str(e))
    return sfs


def _get_default_from_snp_format(filename):
    """
    Returns population labels, the possibility of outgroup and approximation
    of size from file of dadi's SNP format.
    """
    with open(filename) as f:
        line = f.readline()
        while line.startswith('#'):
            line = f.readline()
        # Read the header of file
        info = line.split()
        if (len(info) - 6) % 2 != 0:
            raise ValueError("Cannot calculate number of populations in"
                             " dadi's SNP input file. Maybe it's wrong?")
        n_pop = int((len(info) - 6) / 2)
        pop_ids = info[3: 3 + n_pop]
        # Find approximate size and check existence of the outgroup
        has_outgroup = True
        appr_size = np.zeros(n_pop, dtype=int)
        for line in f:
            info = line.split()
            if info[1][1].lower() not in ['a', 't', 'c', 'g']:
                has_outgroup = False
            for num in range(n_pop):
                cur_size = int(info[3 + num]) + int(info[4 + n_pop + num])
                if cur_size > appr_size[num]:
                    appr_size[num] = cur_size
    return pop_ids, has_outgroup, appr_size


def _change_outgroup(sfs, new_outgroup):
    """
    Change polarization of the data. If data does not have outgroup
    then error.
    """
    if new_outgroup is not None:
        if new_outgroup and sfs.folded:
            raise ValueError("Data does not have outgroup.")
        if not new_outgroup:
            sfs = sfs.fold()
    return sfs


def _read_data_sfs_type(module, data_holder):
    """
    Read filename of dadi's sfs format. Check dadi's manual for further
    information.

    : param module: dadi or moments module (or analogue)
    : param data_holder: object holding the data.
    : type  data_holder: gadma.data.DataHolder
    """
    sfs = module.Spectrum.from_file(data_holder.filename)
    ns = np.array(sfs.shape) - 1

    sfs = _check_missing_population_labels(sfs, data_holder.population_labels,
                                           data_holder.filename)
    sfs = _new_population_labels(sfs, data_holder.population_labels)
    sfs = _project(sfs, data_holder.projections)
    sfs = _change_outgroup(sfs, data_holder.outgroup)
    return sfs


def _read_data_snp_type(module, data_holder):
    """
    Read filename of dadi's SNP format. Check dadi's manual for further
    information.

    : param module: dadi or moments module (or analogue)
    : param data_holder: object holding the data.
    : type  data_holder: gadma.data.DataHolder
    """
    try:
        dd = module.Misc.make_data_dict(data_holder.filename)
    except Exception as e:
        raise SyntaxError("Construction of data_dict failed: " + str(e))
    population_labels, has_outgroup, size = _get_default_from_snp_format(
        data_holder.filename)
    if data_holder.projections is not None:
        size = data_holder.projections
    if data_holder.population_labels is not None:
        population_labels = data_holder.population_labels
    sfs = module.Spectrum.from_data_dict(dd, population_labels,
                                         projections=size,
                                         polarized=has_outgroup)
    if has_outgroup != data_holder.outgroup:
        sfs = _change_outgroup(sfs, data_holder.outgroup)
    return sfs


def read_dadi_data(module, data_holder):
    """
    Read file in one of dadi's formats.

    :param module: dadi or moments module (or analogue)
    :param data_holder: object holding the data.
    :type  data_holder: gadma.data.DataHolder

    :returns: ('sfs'/'snp', data)
    :rtype: (str, :class:`dadi.Spectrum`)
    """
    _, ext = os.path.splitext(data_holder.filename)
    if ext == '.fs' or ext == '.sfs':
        return _read_data_sfs_type(module, data_holder)
    elif ext == '.txt':
        return _read_data_snp_type(module, data_holder)
    else:
        # Try to guess
        try:
            return _read_data_sfs_type(module, data_holder)
        except:  # NOQA
            try:
                return _read_data_snp_type(module, data_holder)
            except:  # NOQA
                raise SyntaxError("Data filename extension is neither .fs"
                                  " (.sfs) or .txt. Attempts to guess the"
                                  " file type failed.\nTo get the error "
                                  "message, please, change the extension.")



import numpy as np

from . import Model
from ..utils import Variable, FractionVariable, float_repr
import copy


class Event(Model):
    """
    Base class for some event.
    """
    def __init__(self):
        super(Event, self).__init__(raise_excep=False)

#    def set_value(self, variable, value):
#        """
#        Fixes variable `variable` to the value of `value`. This variable is
#        no longer available after this operation.
#
#        :param variable: Variable of the event to fix.
#        :param value: New constant value of the variable.
#        """
#        raise NotImplementedError()

    def as_custom_string(self, values):
        """
        Returns string representation of the event.
        """
        raise NotImplementedError()


class Epoch(Event):
    """
    Epoch for demographic model. All arguments could be both values and some
    variables. Additionally values could be combinations of the variables
    (see :class:`gadma.models.VariablesCombinations`).

    :param time_arg: Time of the epoch.
    :type time_arg: float or :class:`gadma.TimeVariable`
    :param init_size_args: Sizes of populations at the beginning of the epoch.
    :type init_size_args: list of values and/or
                          :class:`gadma.PopulationSizeVariable`
    :param size_args: Sizes of populations at the end of the epoch.
    :type size_args: list of values and/or
                     :class:`gadma.PopulationSizeVariable`
    :param mig_args: Migration rates between populations.
    :type mig_args: 2d list of values and/or :class:`gadma.MigrationVariable`
    :param dyn_args: Dynamics of population size changes during the epoch.
    :type dyn_args: list of values and/or  :class:`gadma.DynamicVariable`
    :param sel_args: Selection rates for each population during the epoch.
    :type sel_args: list of values and/or :class:`gadma.SelectionVariable`
    """
    def __init__(self, time_arg, init_size_args, size_args,
                 mig_args=None, dyn_args=None, sel_args=None, dom_args=None):
        # Simple checks
        assert(len(init_size_args) == len(size_args))
        if mig_args is not None:
            mig_args = np.array(mig_args)
            assert(mig_args.ndim == 2)
            assert(len(mig_args) == len(size_args))
            for x in mig_args:
                assert(len(x) == len(size_args))
        if dyn_args is not None:
            assert (len(dyn_args) == len(size_args))
        if sel_args is not None:
            assert (len(sel_args) == len(size_args))
        if dom_args is not None:
            assert (len(dom_args) == len(size_args))

        self.n_pop = len(init_size_args)
        self.time_arg = time_arg
        self.init_size_args = init_size_args
        self.size_args = size_args
        self.sel_args = sel_args
        self.dom_args = dom_args
        if self.dom_args is not None and self.sel_args is None:
            raise ValueError("Dominance coefficients could be set only with"
                             " selection coefficients.")
        self.mig_args = mig_args
        self.dyn_args = dyn_args
        super(Epoch, self).__init__()

        self.add_variable(time_arg)
        self.add_variables(init_size_args)
        self.add_variables(size_args)
        self.add_variables(sel_args)
        self.add_variables(dom_args)
        self.add_variables(dyn_args)
        self.add_variables(mig_args)

#    def set_value(self, variable, value):
#        # check dynamics first as they are more probable in our situation
#        for i, dyn_arg in enumerate(self.dyn_args):
#            if variable is dyn_arg:
#                self.dyn_args[i] = value
#                return
#        if variable is self.time_arg:
#            self.time_arg = value
#        for i, migs in enumerate(self.mig_args):
#            for j, mig_arg in enumerate(migs):
#                if variable is mig_arg:
#                    self.mig_args[i][j] = value
#                return
#        for i, sel_arg in enumerate(self.sel_args):
#            if variable is sel_arg:
#                self.sel_args[i] = value
#                return
#        raise ValueError(f"Event has such variable {variable}. "
#                         f"Available variables: {self.variables}")
#
    def as_custom_string(self, values):
        def _help_f(x, y):
            return f"{y}" if x == "" else f"{y}({x})"

        def help_f(x):
            return _help_f(*self._arg_val_repr(x, values))

        all_repr = [help_f(self.time_arg)]
        sizes_repr = [help_f(arg) for arg in self.size_args]
        sizes_repr = f"[{', '.join(sizes_repr)}]"
        all_repr.append(sizes_repr)
        migs_repr = "[no migs]"
        if self.mig_args is not None:
            migs_repr = [[help_f(mig)for mig in migs]
                         for migs in self.mig_args]
            migs_str = []
            for migs in migs_repr:
                migs_str.append(f"[{', '.join(migs)}]")
            migs_repr = f"[{', '.join(migs_str)}]"
        if self.n_pop > 1:
            all_repr.append(migs_repr)
        sels_repr = ""
        if self.sel_args is not None:
            sels_repr = [help_f(arg) for arg in self.sel_args]
            if self.dom_args is not None:
                dom_repr = [help_f(arg) for arg in self.dom_args]
                together = [f'{x}({y})' for x, y in zip(self.sel_args,
                                                        self.dom_args)]
                sels_repr = f"[{', '.join(together)}]"
            else:
                sels_repr = f"[{', '.join(sels_repr)}]"
            all_repr.append(sels_repr)
        if self.dyn_args is not None:
            dyns_repr = [help_f(arg) for arg in self.dyn_args]
            dyns_repr = f"[{', '.join(dyns_repr)}]"
        else:
            dyns_repr = f"[{', '.join(['Sud' for _ in self.size_args])}]"
        all_repr.append(dyns_repr)
        return f"[ {', '.join(all_repr)} ]"

    def get_vars_not_in_init_args(self):
        variables = copy.copy(self.variables)
        for arg in self.init_size_args:
            if isinstance(arg, Variable):
                if arg in variables:
                    variables.remove(arg)
            elif isinstance(arg, Model):
                rem_vars = []
                for var in variables:
                    if var in arg.variables:
                        rem_vars.append(var)
                for var in rem_vars:
                    variables.remove(var)
        return variables


class Split(Event):
    """
    Class for split demographic event.

    :param pop_to_div: Population index that splits.
    :param size_args: Sizes of populations after split.
    """
    def __init__(self, pop_to_div, size_args=None):
        # Simple checks
        if size_args is not None:
            assert pop_to_div < len(size_args)

        self.n_pop = len(size_args) - 1
        self.pop_to_div = pop_to_div
        self.size_args = size_args
        super(Split, self).__init__()
        self.add_variable(pop_to_div)
        self.add_variables(size_args)

    def as_custom_string(self, values):
        def _help_f(x, y):
            return f"{y}" if x == "" else f"{y}({x.replace(' ', '')})"

        def help_f(x):
            return _help_f(*self._arg_val_repr(x, values))

        frac_str = ""
        for var in self.variables:
            if isinstance(var, FractionVariable):
                val = self.var2value(values)[var]
                frac_str = f" {100 * val : .2f}% ({var.name})"
        sizes_repr = [help_f(arg) for arg in self.size_args]
        sizes_repr = f"[{', '.join(sizes_repr)}]"

        return f"[ {self.pop_to_div + 1} pop split {frac_str} {sizes_repr} ]"



import os
import ruamel.yaml
import numpy as np
from . import settings
from ..data import SFSDataHolder
from ..engines import get_engine, MomentsEngine
from ..models import StructureDemographicModel, CustomDemographicModel
from ..optimizers import get_local_optimizer, get_global_optimizer
from ..optimizers import LinearConstrain
from ..utils import ensure_dir_existence, ensure_file_existence,\
                    check_dir_existence, check_file_existence, abspath,\
                    custom_generator
from ..utils import PopulationSizeVariable, TimeVariable, MigrationVariable

import warnings
import importlib.util
import sys
import copy
import numbers

HOME_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
PARAM_TEMPLATE = os.path.join(HOME_DIR, "params_template")
EXTRA_PARAM_TEMPLATE = os.path.join(HOME_DIR, "extra_params_template")

CHANGED_IDENTIFIERS = {"use_moments_or_dadi": "engine",
                       "size_of_population_in_ga": "size_of_generation",
                       "fractions_in_ga": "fractions",
                       "epsilon": "eps",
                       "stop_iteration": "stuck_generation_number",
                       "name_of_local_optimization": "local_optimizer",
                       "lower_bounds": "lower_bound",
                       "upper_bounds": "upper_bound"}

DEPRECATED_IDENTIFIERS = ["multinom", "verbose", "flush_delay",
                          "epsilon_for_ls", "gtol", "maxiter",
                          "multinomial_mutation", "multinomial_crossing",
                          "distribution", "std",
                          "mean_mutation_rate_for_hc",
                          "const_for_mutation_rate_for_hc",
                          "stop_iteration_for_hc"]


class SettingsStorage(object):
    """
    Class to hold all settings of GADMA run. All default values of settings
    are defined in :mod:`gadma.cli.settings`.
    """
    def __setattr__(self, name, value):
        """
        Sets attribute. The default values of all attributes are in
        :mod:`gadma.cli.settings`. If value is equal to default no checks are
        done. Otherwise value is checked at least for type. If name is not
        known setting then error is raised.

        :param name: Name of attribute.
        :param value: Value of the attribute.
        """
        int_attrs = ['stuck_generation_number', 'sequence_length', 'verbose',
                     'print_models_code_every_n_iteration', 'n_elitism',
                     'draw_models_every_n_iteration', 'size_of_generation',
                     'number_of_repeats', 'number_of_processes',
                     'number_of_populations', 'time_to_print_summary']
        float_attrs = ['theta0', 'time_for_generation', 'eps',
                       'const_of_time_in_drawing', 'vmin', 'min_n', 'max_n',
                       'min_t', 'max_t', 'min_m', 'max_m',
                       'upper_bound_of_first_split',
                       'upper_bound_of_second_split',
                       'const_for_mutation_strength',
                       'const_for_mutation_rate', 'mutation_rate']
        probs_attrs = ['mean_mutation_strength', 'mean_mutation_rate',
                       'p_mutation', 'p_crossover', 'p_random']
        bool_attrs = ['outgroup', 'linked_snp_s', 'only_sudden',
                      'no_migrations', 'silence', 'test', 'random_n_a',
                      'relative_parameters', 'only_models',
                      'symmetric_migrations', 'split_fractions',
                      'generate_x_transform']
        int_list_attrs = ['pts', 'initial_structure', 'final_structure',
                          'projections']
        float_list_attrs = ['lower_bound', 'upper_bound']
        probs_list_attrs = ['fractions']
        attrs_with_equal_len = ['initial_structure', 'final_structure',
                                'population_labels', 'projections']
        special_attrs = ['const_for_mutation_strength',
                         'const_for_mutation_rate', 'vmin',
                         'parameter_identifiers']
        exist_file_attrs = ['input_file', 'custom_filename']
        exist_dir_attrs = ['directory_with_bootstrap', 'resume_from']
        empty_dir_attrs = ['output_directory']
        data_holder_attrs = ['projections', 'outgroup',
                             'population_labels', 'sequence_length']
        bounds_attrs = ['min_n', 'max_n', 'min_t', 'max_t', 'min_m', 'max_m']
        bounds_lists = ['lower_bound', 'upper_bound', 'parameter_identifiers']
        missed_attrs = ['engine', 'local_optimizer', '_inner_data',
                        '_bootstrap_data', 'X_init', 'Y_init', 'model_func',
                        'get_engine_args', 'units_of_time_in_drawing',
                        'data_holder', 'resume_from_settings']

        super_hasattr = True
        setattr_at_the_end = True
        try:
            super(SettingsStorage, self).__getattr__(name)
        except AttributeError:
            super_hasattr = False
        if (name not in int_attrs and name not in float_attrs and
                name not in probs_attrs and name not in bool_attrs and
                name not in attrs_with_equal_len and
                name not in int_list_attrs and
                name not in probs_list_attrs and name not in special_attrs and
                name not in exist_file_attrs and
                name not in exist_dir_attrs and
                name not in empty_dir_attrs and
                name not in data_holder_attrs and
                name not in bounds_attrs and name not in missed_attrs and
                name not in bounds_lists and not super_hasattr):
            raise ValueError(f"Setting {name} should be checked.")

        # -1. For structures it could be one number. We need to transfrom
        # it in list
        if name in int_list_attrs or name in float_list_attrs:
            if isinstance(value, numbers.Real):
                value = [value]
                self.__setattr__(name, value)
                return
        if name == 'population_labels' or name == 'parameter_identifiers':
            if isinstance(value, str):
                if len(value.split(',')) == 1:
                    value = [value.strip()]
                    self.__setattr__(name, value)
                    return

        if isinstance(value, str) and value.lower() == 'none':
            value = None
        # 0. If attribute is equal to the same from setting storage
        # then we let it go any way. It is because of None's in settings
        we_check = True
        if value is None:
            we_check = False
        if hasattr(settings, name):
            default_value = getattr(settings, name)
            if (isinstance(value, np.ndarray) or
                    isinstance(default_value, np.ndarray)):
                if (default_value == value).all():
                    we_check = False
            else:
                if default_value == value:
                    we_check = False
                    try:
                        existed = object.__getattribute__(self, name)
                        if existed != value:
                            delattr(self, name)
                    except AttributeError:
                        pass
                    setattr_at_the_end = False

        # 1. Base checks
        # 1.1 Check is int (positive)
        if name in int_attrs and we_check:
            if (not isinstance(value, numbers.Integral) or
                    isinstance(value, bool)):
                raise ValueError(f"Setting {name} ({value}) must be integer.")
            if value < 0:
                raise ValueError(f"Setting {name} ({value}) must be positive.")
        # 1.2 Check is float and probability
        if (name in float_attrs or name in probs_attrs) and we_check:
            if (not isinstance(value, numbers.Real) or
                    isinstance(value, bool)):
                raise ValueError(f"Setting {name} ({value}) must be float.")
            if name in probs_attrs and (value < 0 or value > 1):
                raise ValueError(f"Setting {name} ({value}) must be between 0 "
                                 "and 1.")
        # 1.3 Check is bool
        if name in bool_attrs and we_check:
            if not isinstance(value, bool):
                raise ValueError(f"Setting {name} ({value}) must be boolean.")
        # 1.4 Check is list of ints
        if name in int_list_attrs and we_check:
            error = ValueError(f"Setting {name} ({value}) must be list of "
                               "integers.")
            if isinstance(value, str):
                try:
                    value = [int(x) for x in value.split(',')]
                except:  # NOQA
                    raise error
            if not isinstance(value, (list, tuple, np.ndarray)):
                raise error
            for val in value:
                if not isinstance(val, numbers.Integral):
                    raise error
                if val < 0:
                    raise ValueError(f"Setting {name} ({value}) have positive"
                                     " elements.")
        # 1.5 Check is the list of floats and probabilities
        if ((name in probs_list_attrs or name in float_list_attrs) and
                we_check):
            if name in probs_list_attrs:
                error = ValueError(f"Setting {name} ({value}) must be list of"
                                   " probabilities.")
            else:
                error = ValueError(f"Setting {name} ({value}) must be list of"
                                   " floats.")
            if isinstance(value, str):
                try:
                    value = [float(x) for x in value.split(',')]
                    self.__setattr__(name, value)
                except:  # NOQA
                    raise error
            if not isinstance(value, (list, tuple, np.ndarray)):
                raise error
            try:
                value = [float(x) for x in value]
            except:  # NOQA
                raise error
            if name in probs_list_attrs:
                for val in value:
                    if val < 0 or val > 1:
                        raise error
        # 1.6 Check that lengths of arrays are equal between lists
        # and equal to the number of populations
        if name in attrs_with_equal_len:
            attrs_list = attrs_with_equal_len
        elif name in bounds_lists:
            attrs_list = bounds_lists
        if ((name in attrs_with_equal_len or name in bounds_lists) and
                we_check):
            if (name not in bounds_lists and
                    hasattr(self, 'number_of_populations') and
                    len(value) != self.number_of_populations):
                raise ValueError(f"Length of {name} should be equal to "
                                 f"{self.number_of_populations}.")
            for attr_name in attrs_list:
                if attr_name == name:
                    continue
                attr_value = getattr(self, attr_name)
                if attr_value is None:
                    continue
                if len(value) != len(attr_value):
                    raise ValueError(f"Setting {name} ({value}) has different"
                                     f" length with another setting "
                                     f"{attr_name} ({attr_value}).")
        # 1.7 Population labels could be taken as one string
        if name == 'population_labels':
            if isinstance(value, str):
                value = [x.strip() for x in value.split(',')]
                self.__setattr__(name, value)

        # 1.8.0 expand path of files:
        if (value is not None and
                (name in empty_dir_attrs or
                 name in exist_file_attrs or name in exist_dir_attrs)):
            value = abspath(value)
        # 1.8 Check for empty dirs (NOW IN ARG_PARSER)
        # if name in empty_dir_attrs and value is not None:
        #     value = ensure_dir_existence(value, check_emptiness=True)
        # 1.9 Check file and dir exist
        if name in exist_file_attrs and value is not None:
            if not check_file_existence(value):
                raise ValueError(f"Setting {name} should be set to existed "
                                 f"file. File {value} does not exist.")
        if name in exist_dir_attrs and value is not None:
            if not check_dir_existence(value):
                raise ValueError(f"Setting {name} should be set to existed "
                                 f"directory. Dir {value} does not exist.")

        # 1.10 Check that identifiers are good:
        if name == "parameter_identifiers" and value is not None:
            # print(name, value)
            if isinstance(value, str):
                value = [x.strip() for x in value.split(",")]
            value = [x.strip() for x in value]
            for val in value:
                if val.lower()[0] not in settings.P_IDS:
                    raise ValueError("Each parameter identifier should start"
                                     " with symbol from the following list: "
                                     f"{settings.P_IDS.keys()}")

        # 2. Dependencies checks
        # 2.1 Const of mutation strength and rate
        if name in ['const_for_mutation_strength', 'const_for_mutation_rate']:
            if value < 1 or value > 2:
                raise ValueError(f"Setting {name} ({value}) must be between 1 "
                                 "and 2.")
        # 2.2 Vmin
        if name == 'vmin':
            if value <= 0:
                raise ValueError(f"Setting {name} ({value}) must be > 0.")

        # 3. Now change some other attributes according to new one
        # 3.1 If new_file we need to create data_holder
        if name == 'input_file':
            data_holder = SFSDataHolder(value,
                                        self.projections,
                                        self.outgroup,
                                        self.population_labels,
                                        self.sequence_length)
            super(SettingsStorage, self).__setattr__('data_holder',
                                                     data_holder)
        # 3.2 If we change some attributes of data_holder we need update it
        elif name in data_holder_attrs:
            if hasattr(self, 'data_holder'):
                setattr(self.data_holder, name, value)
        # 3.3 For engine we need check it exists
        elif name == 'engine':
            engine_obj = get_engine(value)
        # 3.4 For local_optimizer we need check it existence
        elif name == 'local_optimizer':
            optimizer_obj = get_local_optimizer(value)
        # 3.5 If we change engine or pts, we should check for warning if pts
        # would be ignored
        elif name in ['engine', 'pts']:
            if self.engine != 'dadi' and self.pts != settings.pts:
                warnings.warn(f"Engine {self.engine} does not need pts (for "
                              "dadi only). It will be used only for generated"
                              " code for dadi if any.")
        # 3.6 If we set number of populations, we can now check if length of
        # setted attributes are correct. We have already checked that they are
        # equal between each other
        elif name == 'number_of_populations':
            for attr_name in attrs_with_equal_len:
                if getattr(self, attr_name) is None:
                    continue
                if len(getattr(self, attr_name)) != value:
                    raise ValueError(f"Length of {attr_name} should be equal "
                                     f"to {self.number_of_populations}.")
        # 3.7 If we set fractions or size of generation then we create/update
        # GA options that depend on these values.
        elif name == 'fractions' or name == 'size_of_generation':
            fractions = value
            gen_size = value
            if name == 'fractions':
                name = '_fractions'
                if len(fractions) not in [3, 4]:
                    raise ValueError("Length of fractions ({value}) must be "
                                     "equal to 3 (old,mut,cros) or 4 "
                                     "(old,mut,cros,rand). Got length of "
                                     f"{len(values)}")
                if len(fractions) == 3:
                    if sum(value) > 1:
                        raise ValueError('Sum of fractions (when 3 fractions'
                                         ' are setted) must be not greater '
                                         f'than 1. ({value})')
                gen_size = self.size_of_generation
            if name == 'size_of_generation':
                fractions = self.fractions
            if fractions is not None:
                if gen_size is not None:
                    n_elitism = int(fractions[0] * gen_size)
                    self.__setattr__('n_elitism', n_elitism)
                if len(fractions) == 3:
                    p_random = 1 - sum(fractions)
                else:
                    p_random = fractions[3]
                self.__setattr__('p_mutation', fractions[1])
                self.__setattr__('p_crossover', fractions[2])
                self.__setattr__('p_random', p_random)
                if name == '_fractions' and len(value) == 3:
                    value.append(p_random)
        # 3.8 Units of time in drawings
        elif (name == 'units_of_time_in_drawing' or
                name == 'const_of_time_in_drawing'):
            d = {'generations': 1, 'years': 1, 'thousand years': 0.01,
                 'thousands of years': 0.01, 'kya': 0.01}
            if name == 'units_of_time_in_drawing':
                value = value.lower()
                if value not in d:
                    raise ValueError(f"Setting {name} ({value}) must be one "
                                     f"of: {d.keys()}")
                if (value != 'generations' and
                        self.time_for_generation is None):
                    value = 'generations'
                    warnings.warn(f"There is no time for one generation, time"
                                  " will be in generations.")  # TODO move
                object.__setattr__(self, 'const_of_time_in_drawing', d[value])
            else:
                if d[self.units_of_time_in_drawing] != value:
                    found = False
                    for key, val in d.items():
                        if val == value:
                            self.__setattr__('units_of_time_in_drawing', key)
                            found = True
                            break
                    if not found:
                        warnings.warn("No such constant for time drawing. It "
                                      "will be equal to 1 and time will be in"
                                      " generations.")
                        self.__setattr__('units_of_time_in_drawing',
                                         'generations')
                        return

        # 3.9 Domain of variables
        elif name in bounds_attrs:
            if name in ['min_n', 'min_t'] and value <= 0:
                raise ValueError(f"Lower bound {name} should be greater "
                                 "than 0.")
            domain_changed = True
            if name.endswith('n'):
                cls = PopulationSizeVariable
            elif name.endswith('t'):
                cls = TimeVariable
            elif name.endswith('m'):
                cls = MigrationVariable
            else:
                raise AttributeError("Check for supported variables")

            old_domain = np.array(cls.default_domain)

            if name.startswith('min'):
                cls.default_domain[0] = value
            elif name.startswith('max'):
                cls.default_domain[1] = value

            domain_changed = np.any(old_domain != np.array(cls.default_domain))
            if domain_changed:
                if name.endswith('n'):
                    warnings.warn(f"Domain of PopulationSizeVariable changed "
                                  f"to {cls.default_domain}")
                if name.endswith('t'):
                    warnings.warn(f"Domain of TimeVariable changed to "
                                  f"{cls.default_domain}")
                if name.endswith('m'):
                    warnings.warn(f"Domain of MigrationVariable changed to "
                                  f"{cls.default_domain}")

        # 3.10 If we set custom filename with model we should check it is
        # valid python code
        if name == "custom_filename" and value is not None:
            spec = importlib.util.spec_from_file_location("module", value)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            sys.modules['module'] = module
            if hasattr(module, "model_func"):
                func_name = "model_func"
            else:
                raise ValueError("There is no such function `model_func` in "
                                 f" file {value}.")
            model_func_value = getattr(module, func_name)
            self.__setattr__("model_func", model_func_value)

#        if name in bounds_attrs and self.custom_filename is None:
#            msg = f"Setting {name} is set before custom_filename is set."
#            if self.initial_structure is not None:
#                warnings.warn(msg + " It will be ignored as initial structure"
#                              " of model is set.")
#            else:
#                warnings.warn(msg)
        # 3.11 Check for structure or custom filename and ignore some options
        if (name in ['initial_structure', 'final_structure'] and
                value is not None and self.custom_filename is not None):
            if self.lower_bound is not None and self.upper_bound is not None:
                warnings.warn(f"Setting {name} will be ignored as the custom "
                              "model is already set.")

        if ((name == 'lower_bound' and self.upper_bound is not None) or
                (name == 'upper_bound' and self.lower_bound is not None)):
            if name == 'lower_bound':
                lower_bound = value
                upper_bound = self.upper_bound
            else:
                lower_bound = self.lower_bound
                upper_bound = value
            for low, upp in zip(lower_bound, upper_bound):
                if low > upp:
                    raise ValueError(f"Lower bound ({self.lower_bound}) "
                                     f"should be less than upper bound "
                                     f"({self.upper_bound}) element-wise.")
        if setattr_at_the_end:
            super(SettingsStorage, self).__setattr__(name, value)
            # assert(self.__getattr__(name) == value)

    def __getattr__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            # If it is None then maybe we want to return some default value
            if name == "initial_structure" and self.custom_filename is None:
                if hasattr(self, "number_of_populations"):
                    return [settings.initial_structure_unit
                            for _ in range(self.number_of_populations)]
            elif name == "final_structure" and self.custom_filename is None:
                return self.initial_structure
            elif (name == "parameter_identifiers" and
                    self.custom_filename is not None):
                with open(self.custom_filename) as f:
                    big_comment = False
                    big_comment_str = ""
                    found_func = False
                    for line in f:
                        if line.startswith("#") or len(line.strip()) == 0:
                            continue
                        if found_func and line.strip().startswith("'''"):
                            if big_comment and big_comment_str == "'''":
                                big_comment = False
                                big_comment_str = ""
                            else:
                                big_comment = True
                                if big_comment_str == "":
                                    big_comment_str = "'''"
                        elif found_func and line.strip().startswith('"""'):
                            if big_comment and big_comment_str == '"""':
                                big_comment = False
                                big_comment_str = ""
                            else:
                                big_comment = True
                                if big_comment_str == "":
                                    big_comment_str = '"""'
                        elif found_func and not big_comment:
                            break
                        if line.startswith("def model_func"):
                            found_func = True
                    try:
                        p_ids = line.strip().split("=")[0].split(",")
                        p_ids = [x.strip() for x in p_ids]
                        for x in p_ids:
                            settings.P_IDS[x[0].lower()]
                        object.__setattr__(self,
                                           "parameter_identifiers", p_ids)
#                        print(f"Found parameter identifiers in file: {p_ids}")
                        return p_ids
                    except ValueError:  # wrong list
                        pass
                    except IndexError:
                        pass
                    except StopIteration:  # no function
                        pass
                    except KeyError:  # not in P_IDS
                        pass
            elif ((name == "lower_bound" or name == "upper_bound") and
                  (self.custom_filename is not None or
                   self.model_func is not None)):
                if self.parameter_identifiers is not None:
                    bound = list()
                    for p_id in self.parameter_identifiers:
                        domain = settings.P_IDS[p_id[0].lower()].default_domain
                        if name == "lower_bound":
                            bound.append(domain[0])
                        else:
                            bound.append(domain[1])
                    value = bound
                    return bound
            if hasattr(settings, name):
                return getattr(settings, name)
            raise AttributeError(f"There is no such attribute {name} "
                                 "for SettingsStorage.")

    def __eq__(self, other):
        known_missed_attrs = ['data_holder', 'model_func',
                              'units_of_time_in_drawing', 'bootstrap_data',
                              'inner_data', 'number_of_populations',
                              'test', 'generate_x_transform', 'resume_from',
                              'output_directory']
        if not isinstance(other, self.__class__):
            return False
        for attr_name in set(dir(self) + dir(other)):
            if attr_name in known_missed_attrs:
                continue
            if attr_name.startswith("_"):
                continue
            try:
                other_value = getattr(other, attr_name)
            except AttributeError:
                return False
            try:
                self_value = getattr(self, attr_name)
            except AttributeError:
                return False

            if callable(self_value) or callable(other_value):
                continue

            if (isinstance(self_value, np.ndarray) or
                    isinstance(other_value, np.ndarray)):
                if not np.all(np.array(self_value) == np.array(other_value)):
                    return False
            elif (isinstance(self_value, tuple) or
                    isinstance(other_value, tuple)):
                if not list(self_value) == list(other_value):
                    return False
            else:
                if not self_value == other_value:
                    return False
        return True

    @property
    def fractions(self):
        if not hasattr(self, '_fractions'):
            return settings.fractions
        fracs = self._fractions
        if hasattr(self, 'size_of_generation'):
            if self.size_of_generation is not None:
                p_1 = float(self.n_elitism) / self.size_of_generation
            else:
                p_1 = fracs[0]
        fracs = [p_1, self.p_mutation, self.p_crossover, self.p_random]
        return fracs

    def read_data(self):
        """
        Reads data with engine. Attribute of`engine` and `data_holder` should
        be set.
        """
        engine = get_engine(self.engine)
        data = engine.read_data(self.data_holder)
        self.projections = data.sample_sizes
        self.population_labels = data.pop_ids
        self.outgroup = not data.folded  # TODO check function
        if self.pts is None:
            max_n = max(self.projections)
            x = (int((max_n - 1) / 10) + 1) * 10
            super(SettingsStorage, self).__setattr__("pts",
                                                     [x, x + 10, x + 20])
        self._inner_data = data
        self.number_of_populations = len(self.projections)
        return data

    def read_bootstrap_data(self, return_filenames=False):
        """
        Reads all data in the directory `self.bootstrap_data`.

        :param return_filenames: If True then each data is a tuple of
                                 corresponding filename and data.
        """
        if self.directory_with_bootstrap is None:
            return
        if (not hasattr(self, '_inner_data') and
                hasattr(self, 'data_holder') and
                self.data_holder.filename is not None):
            self.read_data()
        dirname = self.directory_with_bootstrap
        engine = get_engine(self.engine)

        all_boot = []
        set_of_seen_files = set()
        filenames = []
        for filename in os.listdir(dirname):
            extention = '.' + filename.split('.')[-1]
            filename_without_ext = '.'.join(filename.split('.')[:-1])
            if filename_without_ext in set_of_seen_files:
                continue
            if not hasattr(self, 'data_holder'):
                data_holder = SFSDataHolder(None)
            else:
                data_holder = copy.deepcopy(self.data_holder)
            data_holder.filename = os.path.join(dirname, filename)
            try:
                data = engine.read_data(data_holder)
                all_boot.append(data)
                if return_filenames:
                    filenames.append(filename)
                set_of_seen_files.add(filename_without_ext)
            except ValueError:
                pass
        self._bootstrap_data = all_boot
        if return_filenames:
            return zip(filenames, all_boot)
        return all_boot

    @property
    def inner_data(self):
        if not hasattr(self, '_inner_data'):
            self._inner_data = self.read_data()
        return self._inner_data

    @property
    def bootstrap_data(self):
        if not hasattr(self, '_bootstrap_data'):
            self._bootstrap_data = self.read_bootstrap_data(self)
        return self._bootstrap_data

    def update_from_file(self, param_file, extra_param_file=None):
        """
        Updates settings by reading new from files.

        :param param_file: File with base parameters.
        :param extra_param_file: File with extra parameters.
        """
        # Load all values
        if param_file is None and extra_param_file is None:
            return self
        loaded_dict = {}
        if param_file is not None:
            with open(param_file) as fl:
                loaded_dict = ruamel.yaml.load(fl,
                                               ruamel.yaml.RoundTripLoader)
        if extra_param_file is not None:
            with open(extra_param_file) as fl:
                extra_dict = ruamel.yaml.load(fl,
                                              ruamel.yaml.RoundTripLoader)
            loaded_dict = {**loaded_dict, **extra_dict}

        for key in loaded_dict:
            attr_name = key.lower().strip()
            attr_name = attr_name.replace(" ", "_")
            attr_name = attr_name.replace("'", "_")
            attr_name = attr_name.replace("__", "_")
#            print(attr_name)
            if attr_name in CHANGED_IDENTIFIERS:
                attr_name = CHANGED_IDENTIFIERS[attr_name]
                setting_name = attr_name.replace("_", " ").capitalize()
                warnings.warn(f"Setting `{key}` is renamed in 2 version of "
                              f"GADMA to `{setting_name}`. It is successfully"
                              " read.")

            if not hasattr(self, attr_name):
                if attr_name in DEPRECATED_IDENTIFIERS:
                    warnings.warn(f"Setting `{key}` was deprecated in 2 "
                                  "version of GADMA. If you have not set it "
                                  "in purpose, ignore this warning.")
                    continue
                else:
                    raise AttributeError(f"Unknown identifier: `{key}`.")
            self.__setattr__(attr_name, loaded_dict[key])
        return self

    @staticmethod
    def from_file(param_file, extra_param_file=None):
        """
        Creates new object with settings from files.

        :param param_file: File with base parameters.
        :param extra_param_file: File with extra parameters.
        """
        obj = SettingsStorage()
        return obj.update_from_file(param_file, extra_param_file)

    def to_files(self, params_file, extra_params_file):
        """
        Saves current options to files.

        :param params_file: File with base parameters.
        :param extra_param_file: File with extra parameters.
        """
        # TODO comments with default values
        known_missed_attrs = ['data_holder', 'model_func',
                              'const_of_time_in_drawing', 'bootstrap_data',
                              'inner_data', 'number_of_populations',
                              'X_init', 'Y_init', 'initial_structure_unit',
                              'test', 'resume_from_settings',
                              'generate_x_transform']
        saved_attrs = []
        for filename, template in zip([params_file, extra_params_file],
                                      [PARAM_TEMPLATE, EXTRA_PARAM_TEMPLATE]):
            with open(template) as fl:
                readed_template = fl.read()
            loaded_template = ruamel.yaml.load(readed_template,
                                               ruamel.yaml.RoundTripLoader)
            for key_name in loaded_template:
                attr_name = key_name.lower()
                attr_name = attr_name.replace("' ", '_')
                attr_name = attr_name.replace(' ', '_')
                attr_name = attr_name.replace("'", '_')
                attr_name = attr_name.replace("__", "_")
                value = getattr(self, attr_name)
                # get rid of numpy as yaml could not serialize it
                if isinstance(value, np.ndarray):
                    value = value.tolist()
                if (isinstance(value, numbers.Integral) and
                        not isinstance(value, bool)):
                    value = int(value)
                elif (isinstance(value, numbers.Real) and
                        not isinstance(value, bool)):
                    value = float(value)
                if isinstance(value, list):
                    for i in range(len(value)):
                        if (isinstance(value[i], numbers.Integral) and
                                not isinstance(value, bool)):
                            value[i] = int(value[i])
                        elif (isinstance(value[i], numbers.Real) and
                                not isinstance(value, bool)):
                            value[i] = float(value[i])

                loaded_template[key_name] = value
                saved_attrs.append(attr_name)

            # For representation of None's and bools
            def my_represent_none(self, data):
                r = self.represent_scalar(u'tag:yaml.org,2002:null', u'Null')
                return r

            def my_represent_bool(self, data):
                if data:
                    value = u'True'
                else:
                    value = u'False'
                return self.represent_scalar(u'tag:yaml.org,2002:bool', value)
            ruamel.yaml.RoundTripRepresenter.add_representer(type(None),
                                                             my_represent_none)
            ruamel.yaml.RoundTripRepresenter.add_representer(bool,
                                                             my_represent_bool)
            with open(filename, 'w') as fl:
                ruamel.yaml.dump(loaded_template, fl,
                                 default_flow_style=True,
                                 Dumper=ruamel.yaml.RoundTripDumper)
        # save missed attributes at the end of extra file
        final_dict = dict()
        with open(extra_params_file, 'a') as fl:
            fl.write("\n#\tOther parameters of run without description:\n")
            for attr_name in set(dir(self) + dir(settings)):
                if (attr_name.startswith("_") or
                        attr_name.isupper() or
                        attr_name in known_missed_attrs or
                        attr_name in saved_attrs):
                    continue
                value = getattr(self, attr_name)
                if callable(value):
                    continue
                # get rid of numpy as yaml could not serialize it
                if isinstance(value, np.ndarray):
                    value = value.tolist()
                final_dict[attr_name] = value
            fl.write(ruamel.yaml.dump(final_dict, default_flow_style=False))

    def get_global_optimizer(self):
        """
        Return object of global optimizer for optimization according to current
        settings.
        """
        # bo = get_global_optimizer("Bayesian_optimization")
        # bo.log_transform = True
        # bo.maximize = True
#        bo = get_global_optimizer("BOTorch")
#        bo.log_transform = True
#        bo.maximize = True
#        return bo
        ga = get_global_optimizer("Genetic_algorithm")
        ga.gen_size = self.size_of_generation
        ga.n_elitism = self.n_elitism
        ga.p_mutation = self.p_mutation
        ga.p_crossover = self.p_crossover
        ga.p_random = self.p_random
        ga.mut_rate = self.mean_mutation_rate
        ga.mut_strength = self.mean_mutation_strength
        ga.const_mut_rate = self.const_for_mutation_rate
        ga.const_mut_strength = self.const_for_mutation_strength
        ga.eps = self.eps
        ga.n_stuck_gen = self.stuck_generation_number
        ga.maximize = True
#        if self.random_n_a:
#            ga.random_type = 'custom'
#            ga.custom_rand_gen = custom_generator
        return ga

    def get_local_optimizer(self):
        """
        Return object of local optimizer for optimization according to current
        settings.
        """
        ls = get_local_optimizer(self.local_optimizer)
        ls.maximize = True
        return ls

#    def get_linear_constrain(self, engine):
#        if (self.upper_bound_of_first_split is None and
#                self.upper_bound_of_second_split is None):
#            return None
#        inv_theta0 = engine.get_N_ancestral_from_theta(1.0)
#
#        A = list()
#        ub = list()
#        if (self.upper_bound_of_first_split is not None):
#            A1, b1 = engine.model.get_involved_for_split_time_vars(1)
#            A.append(A1)
#            ub.append(self.upper_bound_of_first_split /
#                      (2 * inv_theta0) - b1)
#        if (self.upper_bound_of_second_split is not None):
#            A2, b2 = engine.model.get_involved_for_split_time_vars(2)
#            A.append(A2)
#            ub.append(self.upper_bound_of_second_split /
#                      (2 * inv_theta0) - b2)
#        lb = - np.inf * np.ones(len(ub))
#        return LinearConstrainDemographics(np.array(A),
#                                           np.array(lb), np.array(ub),
#                                           engine, self.get_engine_args())

    def get_optimizers_kwargs(self):
        """
        Returns kwargs for optimizations. (`args` and `verbose`).
        """
        kwargs = {}
        kwargs['args'] = self.get_engine_args()
        kwargs['verbose'] = self.verbose
#        kwargs['linear_constrain'] = self.get_linear_constrain()
        return kwargs

    def get_optimizers_init_kwargs(self):
        """
        Returns kwargs for first run of optimization. (`X_init` and `Y_init`).
        """
        return {'X_init': self.X_init, 'Y_init': self.Y_init}

    def get_engine_args(self, engine_id=None):
        """
        Returns `args` of :func:`engine.evaluate` function.
        """
        if engine_id is None:
            engine_id = self.engine
        if engine_id == 'dadi':
            args = (self.pts,)
        else:
            args = (MomentsEngine.default_dt_fac,)
        return args

    def get_linear_constrain_for_model(self, model):
        """
        Returns linear constrain for model based of setted upper bound of
        splits. NOT WORKING.
        """
        if (self.upper_bound_of_first_split is None and
                self.upper_bound_of_second_split is None):
            return None
        A = list()
        ub = list()
        if (self.upper_bound_of_first_split is not None):
            A1, b1 = model.get_involved_for_split_time_vars(1)
            A.append(A1)
            ub.append(self.upper_bound_of_first_split / 2 - b1)
        if (self.upper_bound_of_second_split is not None):
            A2, b2 = model.get_involved_for_split_time_vars(2)
            A.append(A2)
            ub.append(self.upper_bound_of_second_split / 2 - b2)
        lb = - np.inf * np.ones(len(ub))
        return LinearConstrain(np.array(A), np.array(lb), np.array(ub))

    def get_model(self):
        """
        Returns demographic model to use according to current settings.
        """
        gen_time = self.time_for_generation
        theta0 = self.theta0
        mut_rate = self.mutation_rate
        if (self.initial_structure is not None and
                self.final_structure is not None):
            create_migs = not self.no_migrations
            create_sels = False
            create_dyns = not self.only_sudden
            sym_migs = self.symmetric_migrations
            split_f = self.split_fractions
            model = StructureDemographicModel(self.initial_structure,
                                              self.final_structure,
                                              create_migs, create_sels,
                                              create_dyns, sym_migs, split_f,
                                              gen_time, theta0, mut_rate)
            constrain = self.get_linear_constrain_for_model(model)
            model.linear_constrain = constrain
            return model
        elif ((self.custom_filename is not None or
                self.model_func is not None) and
                self.lower_bound is not None and
                self.upper_bound is not None):
            var_classes = list()
            if self.parameter_identifiers is not None:
                for p_id in self.parameter_identifiers:
                    var_classes.append(settings.P_IDS[p_id[0].lower()])
            else:
                for _ in self.lower_bound:
                    var_classes.append(ContinuousVariable)
            names = [f"var{i}" for i in range(len(var_classes))]
            if self.parameter_identifiers is not None:
                p_ids = self.parameter_identifiers
                if len(set(p_ids)) == len(p_ids):
                    names = p_ids
            variables = list()
            for low_bound, upp_bound, var_cls, name in zip(self.lower_bound,
                                                           self.upper_bound,
                                                           var_classes,
                                                           names):
                variables.append(var_cls(name, domain=[low_bound, upp_bound]))
            return CustomDemographicModel(self.model_func, variables,
                                          gen_time, theta0, mut_rate)
        elif self.custom_filename is None and self.model_func is not None:
            return CustomDemographicModel(self.model_func, None,
                                          gen_time, theta0, mut_rate)



import argparse
import os
import sys
import numpy as np
import gadma
import importlib
import copy
import multiprocessing as mp
from importlib.machinery import SourceFileLoader


def run_job(params):
    """
    Run one local search optimization on one bootstrapped data.

    :param params: Tuple of (filename with spectrum, loaded spectrum,
                    initial best values, settings storage).
    """
    boot_fs_filename, boot_fs, p0, settings = params
    np.random.seed()
    fs_filename = gadma.utils.abspath(boot_fs_filename)
    prefix = str(settings.local_optimizer) + '_' + boot_fs_filename
    out_file = os.path.join(settings.output_directory,
                            "opt_" + str(prefix) + ".out")
    res_file = os.path.join(settings.output_directory,
                            str(prefix) + "_params.npy")
    if os.path.isfile(res_file):
        return boot_fs_filename, np.load(res_file, allow_pickle=True)

    kwargs = {'verbose': 1,
              'report_file': out_file}

    custom_model = settings.get_model()

    grid_size = settings.get_engine_args()

    engine = gadma.get_engine(settings.engine)
    engine.set_data(boot_fs)
    engine.set_model(custom_model)

    f = engine.evaluate

    optimizer = gadma.get_local_optimizer(settings.local_optimizer)
    optimizer.maximize = True

    result = optimizer.optimize(f, custom_model.variables, x0=p0,
                                args=grid_size, **kwargs)
    popt = result.x
    theta = engine.get_theta(popt, *grid_size)

    popt = np.append(popt, theta)
    np.save(res_file, popt)
    return (boot_fs_filename, popt)


def load_module(filename):
    """
    Loads module from file.
    """
    filename = gadma.utils.abspath(filename)
    name = "strange_name_of." +\
           os.path.basename(filename).replace('/', '_').rstrip('.py')
    spec = importlib.util.spec_from_loader(name, SourceFileLoader(name,
                                                                  filename))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def load_parameters_from_python_file(filename):
    """
    Loads ``lower_bound``, ``upper_bound``, ``p0`` or ``popt`` and ``pts`` from
    valid python file if they are there.

    :returns: tuple(p0, settings)
    """
    settings = gadma.SettingsStorage()
    if filename is None:
        return None, settings
    try:
        module = load_module(filename)
    except Exception as e:
        raise ValueError('File ' + filename + " is not valid python file." +
                         str(e))
    try:
        settings.lower_bound = module.lower_bound
    except Exception as e:
        settings.lower_bound = None

    try:
        settings.upper_bound = module.upper_bound
    except Exception as e:
        settings.upper_bound = None

    try:
        p0 = module.popt
    except Exception as e:
        try:
            p0 = module.p0
        except Exception as e:
            p0 = None
    try:
        settings.pts = module.pts
    except Exception as e:
        settings.pts = None
    try:
        settings.parameter_identifiers = module.par_labels
    except Exception as e:
        try:
            settings.parameter_identifiers = module.param_labels
        except Exception as e:
            settings.parameter_identifiers = None
    return p0, settings


def main():
    """
    Base function of script parse command-line arguments, loads all
    bootstrapped data and runs local search optimization for each data in
    parallel.

    All output is saved in output directory.
    """
    import pandas as pd
    parser = argparse.ArgumentParser(
        "GADMA module for runs of local search on bootstrapped data. "
        "Is needed for calculating confidence intervals.\n")
    parser.add_argument('-b', '--boots', metavar="<dir>", required=True,
                        help='Directory where bootstrapped data is located.')
    parser.add_argument('-d', '--dem_model', metavar="<filename>",
                        required=True,
                        help="File with demographic model. Should contain "
                             "`model_func` or `generated_model` function. One "
                             "can put there several extra parameters and they "
                             "will be taken automatically, otherwise one will "
                             "need to enter them manually. Such parameters "
                             "are:\n"
                             "\t1) p0 (or popt) - initial parameters values\n"
                             "\t2) lower_bound - list of lower bounds for "
                             "parameters values\n"
                             "\t3) upper_bound - list of pper bounds for "
                             "parameters values\n"
                             "\t4) par_labels/param_labels - list of string "
                             "names for parameters 6) pts - pts for dadi "
                             "(if there is no pts then moments will be run "
                             "automatically).")
    parser.add_argument('-o', '--output', metavar="<dir>", required=True,
                        help='Output directory.')
    parser.add_argument('-j', '--jobs', metavar="N", type=int, default=1,
                        help='Number of threads for parallel run.')

    parser.add_argument('--opt', metavar="log/powell", type=str, default='log',
                        help="Local search algorithm, by now it can be:\n"
                             "\t1) `log` - Inference.optimize_log\n"
                             "\t2) `powell` - Inference.optimize_powell.")
    parser.add_argument('-p', '--params', metavar="<filename>", type=str,
                        default=None,
                        help="Filename with parameters, should be valid "
                             "python file. Parameters are presented in "
                             "-d/--dem_model option description upper.")

    args = parser.parse_args()

    p0, settings = load_parameters_from_python_file(args.dem_model)
    settings.output_directory = args.output
    if args.opt == 'log':
        settings.local_optimizer = 'optimize_log'
    elif args.opt == 'powell':
        settings.local_optimizer = 'optimize_log_powell'
    else:
        settings.local_optimizer = args.opt

    loaded_attrs = ['parameter_identifiers', 'pts',
                    'lower_bound', 'upper_bound']
    if args.params is not None:
        fresh_p0, fresh_settings = load_parameters_from_python_file(
            args.params)

        for attr_name in loaded_attrs:
            if getattr(fresh_settings, attr_name) is not None:
                settings.__setattr__(attr_name,
                                     getattr(fresh_settings, attr_name))
        if fresh_p0 is not None:
            p0 = fresh_p0

    settings.custom_filename = args.dem_model
    settings.get_model()

    for attr_name in loaded_attrs:
        if attr_name != 'pts' and getattr(settings, attr_name) is None:
            raise ValueError(f"Parameter `{attr_name}` is missed both in "
                             f"demographic model and params files")
    if p0 is None:
        raise ValueError(f"Parameter `p0` (or `popt`) is missed both in "
                         "demographic model and params files")

    if settings.pts is None:
        settings.engine = 'moments'
    else:
        settings.engine = 'dadi'
    print(f"Chosen engine: {settings.engine}")

    output = gadma.utils.ensure_dir_existence(args.output)

    settings.directory_with_bootstrap = args.boots
    all_boots = list(settings.read_bootstrap_data(True))

    p_ids = list(settings.parameter_identifiers)
    p_ids.append('Theta')

    print(f"{len(all_boots)} bootstrapped data found.")
    print(f"Found initial parameters values: {p0}")
    print(f"Parameters labels will be: {', '.join(p_ids)}")
    print(f"Lower bound will be: {settings.lower_bound}")
    print(f"Upper bound will be: {settings.upper_bound}")
    print(f"Optimization to run: {settings.local_optimizer}")

    if args.jobs == 1:
        result = []
        for fs_filename, fs in all_boots:
            res = run_job((all_boots[0][0], all_boots[0][1], p0, settings))
            result.append(res)
    else:
        pool = mp.Pool(processes=args.jobs)
        try:
            result = []
            map_result = pool.map_async(run_job,
                                        [(fs_filename, fs, p0, settings)
                                         for fs_filename, fs in all_boots],
                                        callback=result.extend)
            while True:
                try:
                    map_result.get(timeout=1)
                    break
                except mp.TimeoutError as ex:
                    pass
                except Exception as e:
                    pool.terminate()
                    raise e
            pool.close()
            pool.join()
        except KeyboardInterrupt:
            pool.terminate()
            sys.exit(1)

    fs_filenames = [x[0] for x in result]
    result = np.array([x[1] for x in result])
    df = pd.DataFrame(data=result, index=fs_filenames, columns=p_ids)
    csv_path = os.path.join(output, 'result_table.csv')
    pkl_path = os.path.join(output, 'result_table.pkl')
    df.to_csv(csv_path)
    df.to_pickle(pkl_path)
    print(df)
    print(f"DONE. Results are saved as csv file ({csv_path}) and "
          f"as pandas dataframe ({pkl_path}).")



from ..utils import Variable, PopulationSizeVariable, TimeVariable
from ..utils import VariablePool, variables_values_repr
from ..utils import MigrationVariable, DynamicVariable, SelectionVariable
from .demographic_model import DemographicModel
from collections import OrderedDict
import copy
import numpy as np
import importlib.util


class CustomDemographicModel(DemographicModel):
    """
    Demographic model that was defined by the user in some file. Engines
    should be able to deal with such kind of models.

    :param function: function that creates the demographic model or
                     something like that.
    :param variables: Variables of the function. I.e. model parameters.
    :param gen_time: Time of one generation.
    :param theta0: Mutation flux. See :class:`gadma.models.DemographicModel`
                   for more information.
    :param mu: Mutation rate. See :class:`gadma.models.DemographicModel`
               for more information.
    """
    def __init__(self, function, variables,
                 gen_time=None, theta0=None, mu=None):
        self.function = function
        super(CustomDemographicModel, self).__init__(gen_time, theta0, mu)
        if variables is None:
            variables = VariablePool()
        self.variables = variables



from .optimizer import ConstrainedOptimizer
from .global_optimizer import GlobalOptimizer, register_global_optimizer
from .optimizer_result import OptimizerResult
from ..utils import sort_by_other_list, choose_by_weight, eval_wrapper
from ..utils import trunc_normal_3_sigma_rule, DiscreteVariable,\
                    WeightedMetaArray
from ..utils import update_by_one_fifth_rule, ensure_file_existence, fix_args
from ..utils import list_with_weights_for_pickle,\
                    list_with_weights_after_pickle
from functools import partial
import numpy as np
import copy
import pickle
import sys
import time


class GeneticAlgorithm(GlobalOptimizer, ConstrainedOptimizer):
    """
    Class for Genetic Algorithm.

    :param gen_size: Size of generation of genetic algorithm. That is number
                     of individuals/solutions on each step of GA.
    :type gen_size: int
    :param n_elitism: Number of best models from previous generation in GA
                      that will be taken to new iteration.
    :type n_elitism: int
    :param p_mutation: probability of mutation in one generation of GA.
    :type p_mutation: float
    :param p_crossover: probability of crossover in one generation of GA.
    :type p_crossover: float
    :param p_random: Probability of random generated individual in one
                     generation of GA.
    :type p_random: float
    :param mut_rate: Initial mean mutation rate.
    :type mut_rate: float
    :param mut_strength: initial mutation "strength" - mean fraction of
                         model parameters that will be mutated.
    :type mut_strength: float
    :param const_mut_rate: constant to change mutation rate according to
                           one-fifth algorithm. Check GADMA paper for more
                           information.
    :type const_mut_rate: float
    :param eps: const for model's log likelihood compare.
                Model is better if its log likelihood is greater than
                log likelihood of another model by epsilon.
    :type eps: float
    :param n_stuck_gen: Number of iterations for GA stopping: GA stops when
                        it can't improve model during n_stuck_gen generations.
    :type n_stuck_gen: int
    :param selection_type: Type of selection operator in GA. Could be:
                           * 'roulette_wheel'
                           * 'rank'
                           See help(GeneticAlgorithm.selection) for more
                           information.
    :type selection_type: str
    :param selection_random: If True then number of mutants and crossover's
                             offsprings in new generation will be binomial
                             random variable.
    :type selection_type: bool
    :param mutation_type: Type of mutation operator in GA. Could be:
                          * 'uniform'
                          * 'resample'
                          * 'gaussian'
                          See help(GeneticAlgorithm.mutation) for more
                          information.
    :type mutation_type: str
    :param one_fifth_rule: If True then one fifth rule is used in mutation.
    :type one_fifth_rule: bool
    :param crossover_type: Type of crossover operator in GA. Could be:
                           * 'k-point'
                           * 'uniform'
                           See help(GeneticAlgorithm.crossover) for more
                           information.
    :type crossover_type: str
    :param crossover_k: k for 'k-point' crossover type.
    :type crossover_k: int
    :param random_type: Type of random generation of new offsprings. Could be:
                        * 'uniform'
                        * 'resample'
                        * 'custom'
                        See help(GlobalOptimizer.randomize) for more
                        information.
    :type random_type: str
    :param custom_rand_gen: Random generator for 'custom' random_type.
                            Provide generator from variables:
                            custom_rand_gen(variables) = values
    :type custom_rand_gen: func
    :param log_transform: If True then logarithm will be used incide for
                          parameters.
    :type log_transform: bool
    :param maximize: If True then optimization will maximize function.
    :type maximize: bool
    """
    def __init__(self, gen_size=10, n_elitism=2,
                 p_mutation=0.3, p_crossover=0.3, p_random=0.2,
                 mut_strength=0.2, const_mut_strength=1.1,
                 mut_rate=0.2, const_mut_rate=1.2, mut_attempts=2,
                 eps=1e-2, n_stuck_gen=100,
                 selection_type='roulette_wheel', selection_random=False,
                 mutation_type='gaussian', one_fifth_rule=True,
                 crossover_type='uniform', crossover_k=None,
                 random_type='resample', custom_rand_gen=None,
                 log_transform=False, maximize=False):
        # Simple checks
        assert isinstance(gen_size, int)
        assert isinstance(n_elitism, int)
        assert (n_elitism < gen_size)
        assert (p_mutation >= 0 and p_mutation <= 1)
        assert (p_crossover >= 0 and p_crossover <= 1)
        assert (p_random >= 0 and p_random <= 1)
        assert (mut_rate >= 0 and mut_rate <= 1)
        assert (mut_strength >= 0 and mut_strength <= 1)
        assert (const_mut_rate >= 1 and const_mut_rate <= 2)
        assert (const_mut_strength >= 1 and const_mut_strength <= 2)

        self.gen_size = gen_size
        self.n_elitism = n_elitism
        self.p_mutation = p_mutation
        self.p_crossover = p_crossover
        self.p_random = p_random

        self.mut_rate = mut_rate
        self.mut_strength = mut_strength
        self.cur_mut_rate = mut_rate
        self.cur_mut_strength = mut_strength
        self.const_mut_rate = const_mut_rate
        self.const_mut_strength = const_mut_strength
        self.mut_attempts = mut_attempts
        self.eps = eps
        self.n_stuck_gen = n_stuck_gen

        self.selection_type = selection_type
        self.selection_random = selection_random
        self.crossover_type = crossover_type
        self.crossover_k = crossover_k
        self.mutation_type = mutation_type
        self.random_type = random_type
        self.custom_rand_gen = custom_rand_gen
        if self.random_type == 'custom' and self.custom_rand_gen is None:
            raise ValueError("Please specify custom random generator "
                             "(custom_rand_gen) for 'custom' type of random "
                             "sampling.")
        self.one_fifth_rule = one_fifth_rule
        super(GeneticAlgorithm, self).__init__(log_transform, maximize)

    def randomize(self, variables, random_type='resample',
                  custom_rand_gen=None):
        x = super(GeneticAlgorithm, self).randomize(variables, random_type,
                                                    custom_rand_gen)
        x = WeightedMetaArray(x)
        x.metadata = 'r'
        return x

    def mutation_by_ind(self, x, variables, index, mutation_type='gaussian',
                        one_fifth_rule=True):
        """
        Mutation of `x` in index `index`. For more information see
        :meth:`mutation`.
        """
        x_mut = WeightedMetaArray(x, dtype=object)
        var = variables[index]
        # Start mutation procedure
        # 1. Uniform type
        if mutation_type == 'uniform':
            # Case 1 Discrete variable
            if isinstance(var, DiscreteVariable):
                while x[index] == x_mut[index] and len(var.domain) > 1:
                    x_mut[index] = np.random.choice(var.domain)
            # Case 2 Continous variable
            else:
                x_mut[index] = np.random.uniform(var.domain[0], var.domain[1])
        # 2. Resample type
        elif mutation_type == 'resample':
            while x[index] == x_mut[index] and len(var.domain) > 1:
                x_mut[index] = var.resample()
        # 3. Gaussian type
        elif mutation_type == 'gaussian':
            # Choose sign for change
            sign = np.random.choice([-1, 1])
            # Generate mutation rate
            if one_fifth_rule:
                rate = self._sample_mut_rate()
            else:
                rate = self.mut_rate
            # Change value
            # Case 1 Discrete variable
            if isinstance(var, DiscreteVariable):
                i_in_dom = np.where(var.domain == x[index])[0][0]
                new_i = (sign + i_in_dom) % len(var.domain)
                x_mut[index] = var.domain[new_i]
            # Case 2 Continous variable
            else:
                x_mut[index] *= 1 + sign * rate

            # Check if it was something that could be 0
            if var.domain[0] == 0 and sign == -1:
                if x_mut[index] < 1e-5 and np.random.choice([False, True]):
                    x_mut[index] = 0
        else:
            raise ValueError(f"Unknown mutation type: {mutation_type}.")

        # Check x and change weights if they exist
        x_mut = self.check_x(variables, x_mut)
        if isinstance(x, WeightedMetaArray):
            x_mut.weights = copy.deepcopy(x.weights)
            x_mut.metadata = x.metadata
        else:
            x_mut.weights = np.ones(len(x))
        x_mut.weights[index] += 1
        x_mut.metadata += 'm'
        return x_mut

    def mutation(self, x, variables, mutation_type='gaussian',
                 one_fifth_rule=True, attemts=1):
        """
        Mutation operator in genetic algorithm of values `x` of variables
        `variables`. The number of parameters to
        mutate will be sampled from binomial distribution with mean equal to
        mutation strength. The type of change of chosen parameters could be
        set to one of three operators:

        * 'uniform' - new values will be sampled uniformly between bounds.

        * 'resample' - new values will be sampled from the random\
        distribution of the variables.

        * 'gaussian' - will adds a unit Gaussian distributed random value.\
        The mean of the Gaussian distribution will be taken from the mutation\
        rate.

        :param x: values to mutate.
        :param variables: variables.
        :param mutation_type: type of mutation operator. Could be 'gaussian',
                              'uniform' and 'resample'.
        :type mutation_type: str
        :param fval: max number of function evaluations.
        :param one_fifth_rule: If True then one fifth rule will be used. For
                         'gaussian' option only.
        :param weights: weights for parameters in x, the greater weight is
                        the greater probability to change it is.
        :param attemts: number of mutation attemts.
        :type attemts: int

        :returns: a mutated offspring. If `attempts` > 1 then a list of\
                  mutated offsprings. All offsprings have information about\
                  the number of changes of each parameter in `weights`\
                  attribute.
        """
        # Simple checks
        assert len(x) == len(variables)
        assert attemts > 0

        # Generate number of parameters to change
        num_inds = self._sample_number_of_changes(n=len(x))

        # Choose parameters to change according to the weights if they are
        # set otherwise uniformly
        if isinstance(x, WeightedMetaArray):
            weights = x.weights
        else:
            weights = np.ones(num_inds)
        inds = choose_by_weight(range(len(x)), weights, num_inds)

        # Copy the array to change
        x_mut = [np.array(x, dtype=object) for _ in range(attemts)]

        # Start mutation procedure
        for attempt in range(attemts):
            i_att = 0
            while np.all(x_mut[attempt] == x) and i_att < 5:
                i_att += 1
                for ind in inds:
                    x_mut[attempt] = self.mutation_by_ind(x_mut[attempt],
                                                          variables, ind,
                                                          mutation_type,
                                                          one_fifth_rule)
        return x_mut

    def crossover(self, parent1, parent2, variables, crossover_type='uniform',
                  k=2, one_child=True):
        """
        Crossover operator in genetic algorithm. Could be of two types:

        * 'k-point' - k points will be chosen among the vector and each part\
        between those points will be taken from parent1 or parent2 (swapping)\
        . By default k=2.

        * 'uniform' - each parameter will be taken from either parent with\
        equal probability.

        :param parent1: array of first parent.
        :param parent2: array of second parent.
        :param crossover_type: type of crossover operator. Could be 'k-point'
                               or 'uniform'.
        :type crossover_type: str
        :param k: value of k for 'k_point' crossover.
        :type k: int
        :param one_child: if True then one child will be generated and
                          returned.
        :type one_child: bool
        """
        assert len(parent1) == len(parent2)

        parent1 = np.array(parent1, dtype=object)
        parent2 = np.array(parent2, dtype=object)

        # Create two children - copies of parents
        child1 = WeightedMetaArray(parent1, dtype=object)
        child2 = WeightedMetaArray(parent2, dtype=object)

        i_att = 0
        while len(parent1) > 1 and np.all(child1 == parent1) and i_att < 5:
            i_att += 1
            if crossover_type == 'k_point':
                assert k > 0
                assert k < len(parent1)
                # Create list of points
                swp_inds = np.random.choice(range(len(parent1)), size=k)
                # One index must be inside array to make child different to
                # parents
                swp_inds[-1] = np.random.choice(range(1, len(parent1) - 1))
                swp_inds = sorted(swp_inds)
                swp_inds.insert(0, 0)
                swp_inds.append(len(parent1))
                # Swap parameters between points
                for i in range(len(swp_inds) - 1):
                    if i % 2 == 1:
                        p1, p2 = swp_inds[i], swp_inds[i+1]
                        child1[p1:p2] = parent2[p1:p2]
                        child2[p1:p2] = parent1[p1:p2]
            elif crossover_type == 'uniform':
                for i in range(len(parent1)):
                    change = np.random.choice([False, True])
                    if change:
                        child1[i] = parent2[i]
                        child2[i] = parent1[i]
            else:
                raise ValueError(f"Unknown crossover type: {crossover_type}.")

        child1 = self.check_x(variables, child1)
        child2 = self.check_x(variables, child2)
        child1.metadata = 'c'
        child2.metadata = 'c'
        ret = (child1, child2)
        if one_child:
            return ret[np.random.choice([0, 1])]
        return ret

    def selection(self, f, variables, X_gen, Y_gen=None,
                  selection_type='roulette_wheel', selection_random=False):
        """
        Perform selection in genetic algorithm.
        Selection could be of different types:

        * Roulette Wheel - the better fitness function is the higher chance
          to be selected for mutation and crossover for the individual is.
        * Rank - almost the same as Roulette Wheel but with rank insted
          fitness function. This means weight=1 for the best individual,
          weight=2 for the second best and so on.

        :param X_gen: previous generation of individuals.
        :param Y_gen: fitnesses of the previous generation. If `None` then
                         will be evaluated.
        :param selection_type: type of selection. Could be 'roulette_wheel' or
                               'rank'.
        :param selection_random: if True then number of mutants and crossover's
                                 offsprings in new generation will be binomial
                                 random variable.

        :returns: new generation and its fitnesses.
        """
        # Evaluate fitness if None
        if Y_gen is None:
            Y_gen = [f(x) for x in X_gen]
        # Sort by value of fitness
        X_gen, Y_gen = sort_by_other_list(X_gen, Y_gen, reverse=self.maximize)

        # Simple checks
        assert len(X_gen[0]) == len(variables)
        assert len(X_gen) == len(Y_gen)

        # Start selection procedure
        if selection_type == 'roulette_wheel':
            Y_gen = np.array(Y_gen)
            is_not_inf = np.logical_not(np.isinf(Y_gen))
            p = Y_gen / np.sum(Y_gen[is_not_inf])
            p[np.isinf(p)] = 1
            # We need to reverse probs as we have minimization problem
            p = 1 - p
            p /= np.sum(p)
        elif selection_type == 'rank':
            n = len(X_gen)
            p = np.arange(1, n+1) / (n * (n - 1))
            p /= np.sum(p)
        else:
            raise ValueError(f"Unknown selection type: {selection_type}.")

        # Generate numbers for each operation
        if selection_random:
            n_mutants = np.random.binomial(self.gen_size, self.p_mutation)
            n_offsprings = np.random.binomial(self.gen_size, self.p_crossover)
            n_random_gen = np.random.binomial(self.gen_size, self.p_random)
        else:
            n_mutants = int(self.gen_size * self.p_mutation)
            n_offsprings = int(self.gen_size * self.p_crossover)
            n_random_gen = int(self.gen_size * self.p_random)

        # 1. Elitism
        new_X_gen = list(X_gen[:self.n_elitism])
        new_Y_gen = list(Y_gen[:self.n_elitism])

        # 2. Mutation
        for i in range(n_mutants):
            x_ind = np.random.choice(range(len(X_gen)), p=p)
            x = X_gen[x_ind]
            mutants = self.mutation(x, variables, self.mutation_type,
                                    self.one_fifth_rule, self.mut_attempts)
            fitness = [f(x_mut) for x_mut in mutants]

#            print("Time of main part of mutation: " + str(t3 - t1))
            # Take best mutant
            new_Y_gen.append(np.min(fitness))
            new_X_gen.append(mutants[fitness.index(new_Y_gen[-1])])

            # One more check for weights.
            # If new x is better, then we would like to decrease weights of
            # parameters back as this change was good.
            if new_Y_gen[-1] < Y_gen[x_ind]:
                if isinstance(x, WeightedMetaArray):
                    new_X_gen[-1].weights = x.weights

        # 3. Crossover
        for i in range(n_offsprings):
            ind1, ind2 = np.random.choice(range(len(X_gen)), size=2, p=p)
            parent1, parent2 = X_gen[ind1], X_gen[ind2]
            x = self.crossover(parent1, parent2, variables,
                               self.crossover_type, self.crossover_k)
            new_X_gen.append(x)
            new_Y_gen.append(f(x))

        # 4. Random individuals
        for i in range(n_random_gen):
            x = WeightedMetaArray(self.randomize(variables, self.random_type,
                                                 self.custom_rand_gen),
                                  dtype=object)
            x.metadata = 'r'
            new_X_gen.append(x)
            new_Y_gen.append(f(x))

        # Sort by fitness and return new generation
        new_X_gen, new_Y_gen = sort_by_other_list(new_X_gen, new_Y_gen,
                                                  reverse=False)
        new_X_gen = new_X_gen[:self.gen_size]
        new_Y_gen = new_Y_gen[:self.gen_size]
        return new_X_gen, new_Y_gen

    def _sample_mut_rate(self, mode='normal'):
        if mode == 'normal':
            # TODO: Think about std for this distribution
            return trunc_normal_3_sigma_rule(self.cur_mut_rate, 0.0, 1.0)
        elif mode == 'uniform':
            return np.random.uniform(0.0, 1.0)

    def _sample_number_of_changes(self, n):
        sample = np.random.binomial(n=n, p=self.cur_mut_strength)
        return max(1, int(sample))

    def check_x(self, variables, x, raises=False):
        for val, var in zip(x, variables):
            if raises:
                if not var.correct_value(val):
                    raise ValueError("Values in values vector are"
                                     " not in bounds.")
        for i in range(len(x)):
            if not isinstance(variables[i], DiscreteVariable):
                x[i] = min(x[i], variables[i].domain[1])
                x[i] = max(x[i], variables[i].domain[0])
                assert variables[i].correct_value(x[i])
            else:
                if not variables[i].correct_value(x[i]):
                    raise ValueError(f"Value ({x[i]}) of Discrete variable "
                                     f"is bad. Domain: {variables[i].domain}")
        return x

    def is_stopped(self, n_gen, n_eval, impr_gen=None, maxiter=None,
                   maxeval=None, ret_status=False):
        """
        Returns if genetic algorithm must stop.

        :param n_gen: current number of generations.
        :param n_eval: current number of function evaluations.
        :param impr_gen: number of last generation that improved value of
                         fitness function.
        :param maxiter: maximum number of generations.
        :param maxeval: maximum number of evaluation.
        :param ret_status: If True then return status and message.
        """
        status = 1
        message = "OPTIMIZATION IS NOT STOPPED"
        if impr_gen is None:
            impr_gen = n_gen
        is_stuck = (n_gen - impr_gen) >= self.n_stuck_gen
        if maxeval is not None:
            expect_feval = int(self.gen_size * self.p_mutation *
                               self.mut_attempts)\
                         + int(self.gen_size * self.p_crossover)\
                         + int(self.gen_size * self.p_random)
            stop_by_n_eval = (n_eval + expect_feval >= maxeval)
        else:
            stop_by_n_eval = False

        stop_by_n_gen = False
        if maxiter:
            stop_by_n_gen = n_gen >= maxiter

        if stop_by_n_eval:
            status = 1
            message = "MAXIMUM NUMBER OF FUNCTION EVALUATIONS ACHIEVED"
        if is_stuck:
            status = 0
            message = f"CONVERGENCE: NO IMPROVEMENT DURING {self.n_stuck_gen}"\
                      " ITERATIONS"
        ret_value = is_stuck or stop_by_n_eval or stop_by_n_gen
        if ret_status:
            return ret_value, status, message
        return ret_value

    def write_report(self, n_gen, variables, X_gen, Y_gen, x_best, y_best,
                     mean_time, report_file):
        """
        Write report about one generation in report file.

        :note: All values are reported as is, i.e. `X_gen`, `x_best` should be\
               already translated from log scale if optimization did so;\
               `Y_gen` and `y_best` must be already multiplied by -1 if we\
               have maximization instead of minimization.
        """
        if report_file:
            stream = open(report_file, 'a')
        else:
            stream = sys.stdout
        print(f"Generation #{n_gen}.", file=stream)
        print("Current generation of solutions:", file=stream)
        print("N", "Value of fitness function", "Solution",
              file=stream, sep='\t')
        if report_file:
            stream.close()
        for i, (x, y) in enumerate(zip(X_gen, Y_gen)):
            # Use parent's report write function
            super(GeneticAlgorithm, self).write_report(i, variables, x,
                                                       f'{y: 5f}',
                                                       report_file)
        if report_file:
            stream = open(report_file, 'a')

        if self.one_fifth_rule:
            print(f"Current mean mutation rate:\t{self.cur_mut_rate: 3f}",
                  file=stream)
        print(f"Current mean number of params to change during mutation:\t"
              f"{max(int(self.cur_mut_strength * len(x_best)), 1): 3d}",
              file=stream)

        print("\n--Best solution by value of fitness function--", file=stream)
        print("Value of fitness:", y_best, file=stream)
        print("Solution:", file=stream, end='')
        if report_file:
            stream.close()
        super(GeneticAlgorithm, self).write_report('', variables, x_best,
                                                   '', report_file)
        if report_file:
            stream = open(report_file, 'a')

        if mean_time is not None:
            print(f"\nMean time:\t{mean_time:.3f} sec.\n", file=stream)
        print("\n", file=stream)

        if report_file:
            stream.close()

    def save(self, n_gen, n_eval, n_impr_gen, X_gen, Y_gen, X_total, Y_total,
             save_file):
        """
        Save some values of genetic algorithm to file.
        """
        if save_file is None:
            return
        info = (int(n_gen), int(n_eval), int(n_impr_gen),
                list_with_weights_for_pickle(X_gen), Y_gen,
                list_with_weights_for_pickle(X_total), Y_total,
                float(self.cur_mut_rate), float(self.cur_mut_strength))
        super(GeneticAlgorithm, self).save(info, save_file)

    def valid_restore_file(self, save_file):
        try:
            info = self.load(save_file)
        except Exception as e:
            return False
        if not isinstance(info, tuple):
            return False
        if not len(info) == 9:
            return False
        if (not isinstance(info[0], int) or not isinstance(info[1], int) or
                not isinstance(info[2], int)):
            return False
        if not isinstance(info[3], list) or not isinstance(info[4], list):
            return False
        if not isinstance(info[5], list) or not isinstance(info[6], list):
            return False
        if not isinstance(info[7], float) or not isinstance(info[8], float):
            return False
        return True

    def load(self, save_file):
        """
        Load some values of genetic algorithm from file.
        """
        (n_gen, n_eval, n_impr_gen,
         X_gen, Y_gen, X_total, Y_total,
         cur_rate, cur_strength) = super(GeneticAlgorithm,
                                         self).load(save_file)
        return (n_gen, n_eval, n_impr_gen,
                list_with_weights_after_pickle(X_gen), Y_gen,
                list_with_weights_after_pickle(X_total), Y_total,
                cur_rate, cur_strength)

    def optimize(self, f, variables, args=(), num_init=50,
                 X_init=None, Y_init=None,
                 linear_constrain=None, maxiter=None, maxeval=None,
                 verbose=0, callback=None, report_file=None, eval_file=None,
                 save_file=None, restore_file=None, restore_points_only=False,
                 restore_x_transform=None):
        r"""
        Return best values of `variables` that minimizes/maximizes
        the function `f`.

        :param f: function to minimize/maximize. The usage must be the
                  following: f(x, *args), where x is list of values.
        :param variables: list of variables (instances of
                          :class:`gadma.Variable` class) of the function.
        :param X_init: list of initial values.
        :param Y_init: value of function `f` on initial values from `X_init`.
        :param args: arguments of function `f`.
        :param maxiter: maximum number of genetic algorithm's generations.
        :param maxeval: maximum number of function evaluations.
        :param callback: callback to call after each generation.
                         It will be called as callback(x, y), where x, y -
                         best_solution of generation and its fitness.
        """
        # First we initialize initial values of some options
        self.cur_mut_rate = self.mut_rate
        self.cur_mut_strength = self.mut_strength
        # initial number of generation and last impr generation
        n_gen = 0
        n_impr_gen = n_gen
        n_eval_init = 0
        X_total = []
        Y_total = []

        # Create logging files
        if eval_file is not None:
            ensure_file_existence(eval_file)
        if report_file is not None:
            ensure_file_existence(report_file)
        if save_file is not None:
            ensure_file_existence(save_file)

        # Prepare function to use it.
        # Fix args and cache
        prepared_f = self.prepare_f_for_opt(f, args)
        # Wrap for automatic evaluation logging
        finally_wrapped_f = eval_wrapper(prepared_f, eval_file)
        f_in_opt = partial(self.evaluate, finally_wrapped_f)
        f_in_opt = fix_args(f_in_opt, (), linear_constrain)

        # prepare variables
        vars_in_opt = copy.deepcopy(variables)
        for var in vars_in_opt:
            if not isinstance(var, DiscreteVariable):
                var.domain = self.transform(var.domain)

        # Prepare callback if need
#        if callback is not None:
#            callback = self.prepare_callback(callback)

        # Write first line of report
        if verbose > 0 and report_file is not None:
            report_file = ensure_file_existence(report_file)

        # Initial design
        restored = False
        if restore_file is not None and self.valid_restore_file(restore_file):
            (n_gen_old, n_eval_old, n_impr_gen_old, X_gen, Y_gen, X_total,
             Y_total, cur_mut_rate, cur_mut_strength) = self.load(restore_file)
            if restore_x_transform is not None:
                X_gen = [restore_x_transform(x) for x in X_gen]
                X_total = [restore_x_transform(x) for x in X_total]
                # Let's remember our recalculation but this Y_gen will become
                # Y_init and Y_init is in usual units without * sign.
                Y_gen = [self.sign * f_in_opt(x) for x in X_gen]
                Y_total = [None for _ in X_total]
            if not restore_points_only:
                n_gen = n_gen_old
                n_impr_gen = n_impr_gen_old
                self.cur_mut_rate = cur_mut_rate
                self.cur_mut_strength = cur_mut_strength
                n_eval_init = n_eval_old
                num_init = len(X_gen)
            if X_init is None:
                X_init = X_gen
                Y_init = Y_gen
            elif Y_init is None:
                if Y_gen is not None:
                    assert len(X_gen) == len(Y_gen)
                    Y_gen.extend(Y_init)
                X_gen.extend(X_init)
                X_init, Y_init = X_gen, Y_gen
            elif len(X_init) == len(Y_init):
                X_init.extend(X_gen)
                Y_init.extend(Y_gen)
            else:
                assert len(X_init) > len(Y_init)
                new_X_init = copy.copy(X_init[:len(Y_init)])
                new_X_init.extend(X_gen)
                new_X_init.extend(X_init[len(Y_init):])
                Y_init.extend(Y_gen)
                X_init = new_X_init
            restored = True

        # Perform 0 generation of GA - initial design.
        if X_init is not None:
            X_init = [self.transform(x) for x in X_init]
        if Y_init is not None:
            Y_init = [self.sign * y for y in Y_init]
        X_gen, Y_gen = self.initial_design(f_in_opt, variables, num_init,
                                           X_init, Y_init, self.random_type,
                                           self.custom_rand_gen)
        X_gen, Y_gen = sort_by_other_list(X_gen, Y_gen, reverse=False)
        X_gen = X_gen[:self.gen_size]
        Y_gen = Y_gen[:self.gen_size]

        # transform for save function and result
        # X_gen and Y_gen are in translated form and X_gen_cor, Y_gen_cor not
        # x_best and y_best will be in good units!!
        X_gen_cor = [self.inv_transform(x) for x in X_gen]
        Y_gen_cor = [self.sign * y for y in Y_gen]

        # Save total
        X_total.extend(copy.deepcopy(X_gen_cor))
        Y_total.extend(copy.deepcopy(Y_gen_cor))

        # Initialize number of generations, evaluations, best values and so on
        # x_best and y_best will be in good units!!
        n_eval = n_eval_init + prepared_f.cache_info.misses
        x_best = X_gen_cor[0]
        y_best = Y_gen_cor[0]
        assert restored or n_eval > 0

        # Write report about 0 generation
        # we save and report in units of function f that we have got.
        if verbose > 0:
            self.write_report(n_gen, variables, X_gen_cor, Y_gen_cor,
                              x_best, y_best, None, report_file)
        self.save(n_gen, n_eval, n_impr_gen, X_gen_cor, Y_gen_cor,
                  X_total, Y_total, save_file)

        # keep times of generation evaluations
        times = list()

        # Begin to create generations
        while (len(variables) > 0 and
                not self.is_stopped(n_gen, n_eval, n_impr_gen,
                                    maxiter, maxeval)):
            # record time of generation start
            start_time = time.time()
            # Form new generation
            X_gen, Y_gen = self.selection(f_in_opt, variables, X_gen, Y_gen,
                                          self.selection_type,
                                          self.selection_random)

            X_gen_cor = [self.inv_transform(x) for x in X_gen]
            Y_gen_cor = [self.sign * y for y in Y_gen]

            # Save all generations.
            X_total.extend(copy.deepcopy(X_gen_cor))
            Y_total.extend(copy.deepcopy(Y_gen_cor))

            # Check if we improve the result
            if self.sign * (y_best - Y_gen_cor[0]) >= self.eps:
                n_impr_gen = n_gen
                x_best = X_gen_cor[0]
                y_best = Y_gen_cor[0]

            # Update mutation rates and strength
            is_impr = (n_impr_gen == n_gen)
            if self.one_fifth_rule:
                self.cur_mut_rate = update_by_one_fifth_rule(
                    self.cur_mut_rate, self.const_mut_rate, is_impr)
                self.cur_mut_rate = min(self.cur_mut_rate, 1.0)
            is_mut_best = False
            if hasattr(x_best, 'weights') and len(x_best.metadata) > 0:
                is_mut_best = x_best.metadata[-1] == 'm'
            self.cur_mut_strength = update_by_one_fifth_rule(
                self.cur_mut_strength, self.const_mut_strength,
                is_impr and is_mut_best)
            self.cur_mut_strength = min(self.cur_mut_strength, 1.0)

            # Update numbers
            n_gen += 1
            n_eval = n_eval_init + prepared_f.cache_info.misses
            times.append(time.time() - start_time)

            # Callback
            if callback is not None:
                if n_impr_gen == n_gen - 1:
                    callback(x_best, y_best)

            # Write report about current generation
            if verbose > 0 and n_gen % verbose == 0:
                self.write_report(n_gen, variables, X_gen_cor, Y_gen_cor,
                                  x_best, y_best, np.mean(times), report_file)
            # Save generation
            self.save(n_gen, n_eval, n_impr_gen, X_gen_cor, Y_gen_cor,
                      X_total, Y_total, save_file)

        # Construct OptimizerResult object to return
        _, status, message = self.is_stopped(n_gen, n_eval, n_impr_gen,
                                             maxiter, maxeval,
                                             ret_status=True)
        result = OptimizerResult(x=x_best, y=y_best, success=True,
                                 status=status, message=message, X=X_total,
                                 Y=Y_total, n_eval=n_eval, n_iter=n_gen,
                                 X_out=X_gen_cor, Y_out=Y_gen_cor)
        return result


register_global_optimizer('Genetic_algorithm', GeneticAlgorithm)



import sys
from functools import wraps
import numpy as np
import copy

from .global_optimizer import GlobalOptimizer
from .optimizer import ConstrainedOptimizer
from .optimizer_result import OptimizerResult
from ..utils import DiscreteVariable, sort_by_other_list


class GlobalOptimizerAndLocalOptimizer(GlobalOptimizer, ConstrainedOptimizer):
    """
    Class for run of global optimizer followed by local optimizer. It is
    classified as global optimizer with constrains. Function :meth:`optimize`
    got function and variables, runs global optimizer then filter discrete
    variables out and optimize the rest with local optimizer.

    :param global_optimizer: Global optimizer to use inside.
    :type global_optimizer: :class:`gadma.optimizers.GlobalOptimizer`
    :param local_optimizer: Local optimizer to use inside.
    :type local_optimizer: :class:`gadma.optimizers.LocalOptimizer`
    """
    def __init__(self, global_optimizer, local_optimizer):
        self.global_optimizer = global_optimizer
        self.local_optimizer = local_optimizer
        if self.global_optimizer.maximize != self.local_optimizer.maximize:
            raise ValueError("Global and local optimizers must both maximize "
                             "(or minimize) the function.")
        # For restore and save files
        if not hasattr(self.global_optimizer, "id"):
            self.global_optimizer.id = "global_optimizer"
        if not hasattr(self.local_optimizer, "id"):
            self.local_optimizer.id = "local_optimizer"

        super(GlobalOptimizerAndLocalOptimizer, self).__init__(
            log_transform=False, maximize=self.global_optimizer.maximize)

    @staticmethod
    def _get_filter(variables):
        """
        Returns filter to filter out discrete variables for local optimizer.
        Returns numpy array of bools.

        :param variables: List of Variables to get filter for.
        """
        is_not_discrete = [not isinstance(var, DiscreteVariable)
                           for var in variables]
        return np.array(is_not_discrete, dtype=bool)

    def _transform_x(self, x, base_x, is_filtered):
        """
        Transforms vector `x` back from local optimizer according to filter.

        :param x: Vector from local optimizer.
        :param base_x: Base vector from global optimizer that have fixed values
                       that were filtered out.
        :param is_filtered: Filter that was used on `base_x`.
        """
        full_x = np.array(base_x)
        full_x[is_filtered] = x
        return full_x

    def _f_for_local_opt(self, f, variables, x_best):
        """
        Transforms objective function to take vectors from local optimization.
        """
        is_filtered = self._get_filter(variables)

        @wraps(f)
        def wrapper(x, *args):
            full_x = self._transform_x(x, x_best, is_filtered)
            return f(full_x, *args)
        return wrapper

    def _callback_for_local_opt(self, callback, variables, x_best):
        """
        Transforms callback function to take vectors from local optimization.
        """
        if callback is None:
            return None
        is_filtered = self._get_filter(variables)

        @wraps(callback)
        def wrapper(x, y):
            full_x = self._transform_x(x, x_best, is_filtered)
            return callback(full_x, y)
        return wrapper

    def optimize(self, f, variables, args=(), global_num_init=50,
                 X_init=None, Y_init=None, local_options={},
                 linear_constrain=None,
                 global_maxiter=None, local_maxiter=None,
                 global_maxeval=None, local_maxeval=None,
                 verbose=0, callback=None, eval_file=None,
                 report_file=None, save_file=None,
                 restore_file=None, restore_points_only=False,
                 global_x_transform=None, local_x_transform=None):
        """
        :param f: Objective function.
        :type f: func
        :param variables: List of objective function variables.
        :type variables: list of class:`gadma.utils.VariablesPool`
        :param args: Arguments of `f`.
        :type args: tuple
        :param global_num_init: Number of initial points for global optimizer.
        :type global_num_init: int
        :param X_init: List of initial vectors.
        :type X_init: list
        :param Y_init: List of values of target function on points of `X_init`.
        :type Y_init: list
        :param local_options: Options for local optimizer.
        :type local_options: dict
        :param linear_constrain: Linear constrain on variables.
        :type linear_constrain: :class:`gadma.optimizers.LinearConstrain`
        :param global_maxiter: Maximum number of global optimizer iterations
                               to run.
        :type global_maxiter: int
        :param global_maxeval: Maximum number of function evaluation during
                               global optimization.
        :type global_maxeval: int
        :param local_maxiter: Maximum number of local optimizer iterations
                              to run.
        :type local_maxiter: int
        :param local_maxeval: Maximum number of function evaluation during
                              local optimization.
        :type local_maxeval: int
        :param verbose: Varbosity of reports. If 0 then no output.
        :type verbose: int
        :param callback: callback to run after each iteration of both
                         optimizers.
        :type callback: func
        :param eval_file: File to save of objective function evaluations.
        :type eval_file: str
        :param report_file: File to save report each `verbose` iteration. If
                            None and `verbose` > 0 then report will be printed
                            to stdout.
        :type report_file: str
        :param save_file: File to save information during the run.
        :type save_file: str
        :param restore_file: File to restore previous run that was saved by
                             :meth:`save` method.
        :type restore_file: str
        :param restore_points_only: Restore run last results and run again from
                                    it.
        :type restore_points_only: bool
        :param global_x_transform: Transformation of vectors after restore
                                   before run of global optimizer.
        :type global_x_transform: func
        :param local_x_transform: Transformation of vectors after restore
                                  before run of local optimizer.
        :type local_x_transform: bool
        """
        if report_file:
            stream = open(report_file, 'a')
        else:
            stream = sys.stdout
        if verbose != 0:
            print(f"--Start global optimization {self.global_optimizer.id}--",
                  file=stream)
        if report_file:
            stream.close()

        # Run global optimizer
        global_result = self.global_optimizer.optimize(f, variables,
                                                       args, global_num_init,
                                                       X_init, Y_init,
                                                       linear_constrain,
                                                       global_maxiter,
                                                       global_maxeval,
                                                       verbose, callback,
                                                       report_file, eval_file,
                                                       save_file,
                                                       restore_file,
                                                       restore_points_only,
                                                       global_x_transform)
        if report_file:
            stream = open(report_file, 'a')
        else:
            stream = sys.stdout
        if verbose != 0:
            print(f"--Finish global optimization {self.global_optimizer.id}--",
                  file=stream)
            print("Result:\n", global_result, file=stream)

        # Transform best x to local optimizer as x0 and functions for local
        x_best = np.array(global_result.x)
        is_not_discrete = self._get_filter(variables)
        x0 = x_best[is_not_discrete]

        variables_local = np.array(variables)[is_not_discrete]

        f_local = self._f_for_local_opt(f, variables, x_best)
        callback_local = self._callback_for_local_opt(callback, variables,
                                                      x_best)
        if verbose != 0:
            print(f"--Start local optimization {self.local_optimizer.id}--",
                  file=stream)

        if report_file:
            stream.close()

        # Run local optimizer
        local_result = self.local_optimizer.optimize(f_local, variables_local,
                                                     x0, args, local_options,
                                                     linear_constrain,
                                                     local_maxiter,
                                                     local_maxeval,
                                                     verbose, callback_local,
                                                     eval_file, report_file,
                                                     save_file,
                                                     restore_file,
                                                     restore_points_only,
                                                     local_x_transform)
        # Create result
        success = local_result.success
        message = f"GLOBAL OPTIMIZATION: {global_result.message}; "\
                  f"LOCAL OPTIMIZATION: {local_result.message}"
        status = local_result.status

        y_best = local_result.y
        x_best = global_result.x
        x_best[is_not_discrete] = local_result.x
        ga_maximize = self.global_optimizer.maximize
        X_out, Y_out = sort_by_other_list(global_result.X_out,
                                          global_result.Y_out,
                                          reverse=ga_maximize)
        if np.allclose(Y_out[0], global_result.y):
            X_out[0] = x_best
        else:
            X_out.insert(0, x_best)
            Y_out.insert(0, y_best)

        X_total = global_result.X + local_result.X
        Y_total = global_result.Y + local_result.Y
        n_eval = global_result.n_eval + local_result.n_eval
        n_iter = global_result.n_iter + local_result.n_iter
        result = OptimizerResult(x_best, y_best, success, status, message,
                                 X_total, Y_total, n_eval, n_iter,
                                 X_out, Y_out)

        if report_file:
            stream = open(report_file, 'a')
        else:
            stream = sys.stdout
        if verbose != 0:
            print(f"--Finish local optimization {self.local_optimizer.id}--",
                  file=stream)
            print("Result:\n", result, file=stream)
        if report_file:
            stream.close()
        return result



import numpy as np
import pandas as pd
import sys
import argparse
from scipy import stats
from .utils import abspath, check_file_existence


def main():
    """
    Main function of script. Reads command-line arguments, reads saved table
    and calculate confidence intervals from it.
    """
    parser = argparse.ArgumentParser("GADMA module for calculating confidence "
                                     "intervals from the result table of local"
                                     " search runs on bootstrapped data.")
    parser.add_argument('input_filename', metavar='<filename>',
                        help="Filename (.csv or .pkl) with result from local "
                             "search runs on bootstrapped data. Output of "
                             "gadma-run_ls_on_boot_data.")
    parser.add_argument('--log', required=False, action='store_true',
                        help="If log then logarithm will be used to calculate "
                             "confidence intervals.")
    parser.add_argument('--tex', required=False, action='store_true',
                        help="LaTex output.")
    parser.add_argument('--acc', required=False, metavar='N', type=int,
                        default=5, help="Precision of an output (dafault: 5).")

    args = parser.parse_args()
    filename = abspath(args.input_filename)
    if not check_file_existence(filename):
        raise ValueError(f"Input file ({filename}) does not exist.")
    ext = filename.split('.')[-1]
    if ext == 'csv':
        df = pd.read_csv(filename, index_col=0)
    elif ext == 'pkl':
        df = pd.read_pickle(filename)
    else:
        raise ValueError(f"Unknown extension of input file ({filename}). "
                         "Valid extension are: .csv and .pkl")

    a = df.values
    if args.log:
        a = np.log(a)

    means = np.mean(a, axis=0)
    stds = np.std(a, axis=0)

    for m, s, x, par_name in zip(means, stds, a.T, df.columns):
        low = m - 1.96 * s
        upp = m + 1.96 * s
        if args.log:
            low = np.exp(low)
            upp = np.exp(upp)
        k2, p = stats.normaltest(x)
        if p < 0.05:
            normtest_str = 'data looks '
            if args.log:
                normtest_str += 'log-normal (fail to reject H0)'
            else:
                normtest_str += 'normal (fail to reject H0)'
        else:
            normtest_str = 'data does not look '
            if args.log:
                normtest_str += 'log-normal (reject H0)'
            else:
                normtest_str += 'normal (reject H0)'
        normtest_str += f' p-value={p:.2e}'
        if args.tex:
            format_string = '%s:\t$[%.' + str(args.acc) + 'f - %.' +\
                            str(args.acc) + 'f]$\t%s'
        else:
            format_string = '%s:\t%.' + str(args.acc) + 'f\t%.' +\
                            str(args.acc) + 'f\t%s'
        print(format_string % (par_name.strip(), low, upp, normtest_str))



from ..data import DataHolder
from ..utils import Variable
from ..models import BinaryOperation
import copy

_registered_engines = {}


def register_engine(engine):
    """
    Registers the specified engine of the demographic inference.

    :raises ValueError: if engine with the same ``id`` was already\
                        registered.
    """
    if engine.id in _registered_engines:
        raise ValueError(f"Engine of the demographic inference '{engine.id}'"
                         f" already registered.")
    _registered_engines[engine.id] = engine


def get_engine(id):
    """
    Returns the engine of the demographic inference with the specified id.

    :raises ValueError: if engine with such ``id`` was not registered.
    """
    if id not in _registered_engines:
        raise ValueError(f"Engine of the demographic inference '{id}'"
                         f" not registered")
    return _registered_engines[id]()


def all_engines():
    """
    Returns an iterator over all registered engines of the demographic
    inference.
    """
    for engine in _registered_engines.values():
        yield engine()


class Engine(object):
    """
    Abstract class representing an engine of the demographic inference.

    New engine should be inheritted from this class.
    Engine must have at least the ``id``, ``supported_models``,
    ``supported_data`` and ``inner_data`` attributes, and
    implementations of :func:`read_data`
    and :func:`evaluate` functions of this abstract class.

    :cvar str Engine.id: the unique identifier of the engine.
    :cvar Engine.supported_models: list of supported :class:`Model` classes.
    :cvar Engine.supported_data: list of supported :class:`DataHolder` classes.
    :cvar Engine.inner_data_type: class of inner data that is used by engine.
    """
    id = ''
    supported_models = []
    supported_data = []
    inner_data_type = None

    def __init__(self, data=None, model=None):
        self.data = data
        self.model = model

    @staticmethod
    def get_value_from_var2value(var2value, entity):
        if isinstance(entity, Variable):
            return var2value[entity]
        if isinstance(entity, BinaryOperation):
            return entity.get_value(var2value)
        return entity

    @staticmethod
    def read_data(data_holder):
        """
        Reads data from `data_holder.filename` in inner type.

        :param data_holder: Holder of data to read.
        :type data_holder: :class:`gadma.DataHolder`

        :returns: readed data
        :rtype: ``Engine.inner_data_type``
        """
        raise NotImplementedError

    def set_model(self, model):
        """
        Sets new model for the engine.

        :param model: new model.
        :type model: :class:`Model`

        :raises ValueError: when model is not supported by engine.
        """
        if model is None:
            self._model = None
            # in moments and dadi it is used to save thetas
            self.saved_add_info = {}
            return
        is_supported = False
        for cls in self.supported_models:
            if issubclass(model.__class__, cls):
                is_supported = True
        if not is_supported:
            raise ValueError(f"Model {model.__class__} is not supported "
                             f"by {self.id} engine.\nThe supported models"
                             f" are: {self.supported_models}.")
        self._model = model
        self.saved_add_info = {}

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        self.set_model(model)

    def set_data(self, data):
        """
        Sets new data for the engine.

        :param data: new data.
        :type data: :class:`DataHolder` or ``inner_data_type``.

        :raises ValueError: when ``data`` is not supported by the engine.
        """
        if data is None:
            self.data_holder = None
            self.inner_data = None
            self.saved_add_info = {}
            return
        cls = data.__class__
        if cls not in self.supported_data and \
                not isinstance(data, self.inner_data_type):
            try:
                transformed_data = self.inner_data_type(data)
                self.set_data(transformed_data)
            except:  # NOQA
                raise ValueError(f"Data class {cls} is "
                                 f"not supported by {self.id} engine.\n"
                                 f"The supported classes are: "
                                 f"{self.supported_data} and "
                                 f"{self.inner_data_type}")
        if isinstance(data, DataHolder):
            self.inner_data = self.read_data(data)
            self.data_holder = copy.deepcopy(data)
            # TODO valid for dadi and moments
            self.data_holder.projections = self.inner_data.sample_sizes
            self.data_holder.population_labels = self.inner_data.pop_ids
            self.data_holder.outgroup = not self.inner_data.folded
        elif isinstance(data, self.inner_data_type):
            self.inner_data = data
            self.data_holder = None
        self.saved_add_info = {}

    @property
    def data(self):
        return self.inner_data

    @data.setter
    def data(self, new_data):
        self.set_data(new_data)

    def evaluate(self, values, **options):
        """
        Evaluation of the objective function of the engine.

        :param values: values of variables of setted demographic model.
        """
        raise NotImplementedError

    def set_and_evaluate(self, values, model, data, options={}):
        """
        Sets model and data for the engine instance and evaluates the
        objective function via calling :func:`evaluate`.

        :param values: values of variables of the demographic model.
        :param model: model.
        :type model: class from :attr:``supported_models``
        :param data: holder of the data or raw data for the engine.
        :type data: :class:`gadma.DataHolder` or :attr:``inner_data``

        :raises ValueError: if `model` is `None` and any was not\
            setted before; or if `data_holder` is `None` and any\
            was not setted before.
        """
        if model is not None:
            self.set_model(model)
        if data is not None:
            self.set_data(data)
        if self.model is None:
            raise ValueError("Please set model for engine or pass it as "
                             "argument to function.")
        if self.data is None:
            raise ValueError("Please set data for engine or pass it as "
                             "argument to function.")

        return self.evaluate(values, **options)

    def generate_code(self, values, filename=None):
        """
        Prints nice formated code in the format of engine to file or returns
        it as string if no file is set.

        :param values: values for the engine's model.
        :param filename: file to print code. If None then the string will
                         be returned.
        """
        raise NotImplementedError



from ..utils import Variable, PopulationSizeVariable, TimeVariable
from ..utils import VariablePool
from ..utils import MigrationVariable, DynamicVariable, SelectionVariable,\
                    FractionVariable
from . import Model, Epoch, Split
from .variables_combinations import Multiplication, Subtraction
from .demographic_model import EpochDemographicModel
from collections import OrderedDict
import copy
import numpy as np


class StructureDemographicModel(EpochDemographicModel):
    """
    Special class for demographic model created by structure.

    :param initial_structure: List of ints with number of intervals
                              in initial structure.
    :type initial_structure: list of ints
    :param final_structure: List of ints with number of intervals
                            in final structure.
    :type final_structure: list of ints
    :param have_migs: If True then model will have migrations.
    :type have_migs: bool
    :param have_sels: If True then model will have selection coefficients.
    :type have_sels: bool
    :param have_dyns: If True then model will create dynamics of size
                      change different to Sudden change.
    :type have_dyns: bool
    :param sym_migs: If True then migrations will be symetric.
    :type sym_migs: bool
    :param frac_split: If True then populations split in some proportion. If
                       False then newly formed population has size as
                       an independent variable.
    :type frac_split: bool
    """
    def __init__(self, initial_structure, final_structure,
                 have_migs, have_sels, have_dyns, sym_migs, frac_split,
                 gen_time=None, theta0=None, mu=None):
        super(StructureDemographicModel, self).__init__(gen_time, theta0, mu)
        if np.any(np.array(initial_structure) > np.array(final_structure)):
            raise ValueError(f"Elements of the initial structure "
                             f"({initial_structure}) "
                             f"could not be greater than elements of the "
                             f"final structure ({final_structure}).")
        self.initial_structure = np.array(initial_structure)
        self.final_structure = np.array(final_structure)
        self.have_migs = have_migs
        self.have_sels = have_sels
        self.have_dyns = have_dyns
        self.sym_migs = sym_migs
        self.frac_split = frac_split
        self.from_structure(self.initial_structure)

    def from_structure(self, structure):
        """
        Creates new model from given structure.
        It is base constructor of the model.

        :param structure: Structure of the model.
        :type structure: list of ints
        """
        super(StructureDemographicModel, self).__init__(self.gen_time,
                                                        self.theta0, self.mu)
        if not (np.all(np.array(structure) >= self.initial_structure)):
            raise ValueError(f"Elements of model structure ({structure}) "
                             f"could not be smaller than elements of the "
                             f"initial structure ({self.initial_structure}).")

        if not (np.all(np.array(structure) <= self.final_structure)):
            raise ValueError(f"Elements of model structure ({structure}) "
                             f"could not be greater than elements of the "
                             f"final structure ({self.final_structure}).")
        if np.any(np.array(structure) <= 0):
            raise ValueError(f"Elements of model structure ({structure}) "
                             "should be positive (> 0).")

        if list(structure) == [1]:
            return self
        i_int = 0
        if self.frac_split:
            size_vars = [1.0]
        else:
            size_vars = [PopulationSizeVariable('nu')]
        for n_pop in range(1, len(structure) + 1):
            n_int = structure[n_pop - 1]
            if n_pop == 1:
                n_int -= 1
            for _i in range(1, n_int + 1):
                i_int += 1
                time_var = TimeVariable('t%d' % (i_int))
                size_vars = list()
                for i_pop in range(n_pop):
                    var = PopulationSizeVariable('nu%d%d' % (i_int, i_pop+1))
                    size_vars.append(var)
                mig_vars = None
                if self.have_migs and n_pop > 1:
                    mig_vars = np.zeros(shape=(n_pop, n_pop), dtype=object)
                    for i in range(n_pop):
                        for j in range(n_pop):
                            if i == j:
                                continue
                            if self.sym_migs and j < i:
                                continue
                            var = MigrationVariable('m%d_%d%d' %
                                                    (i_int, i+1, j+1))
                            mig_vars[i][j] = var
                            if self.sym_migs:
                                mig_vars[j][i] = var
                sel_vars = None
                if self.have_sels:
                    sel_vars = list()
                    for i in range(n_pop):
                        var = SelectionVariable('g%d%d' % (i_int, i+1))
                        sel_vars.append(var)
                dyn_vars = None
                if self.have_dyns:
                    dyn_vars = list()
                    for i in range(n_pop):
                        var = DynamicVariable('dyn%d%d' % (i_int, i+1))
                        dyn_vars.append(var)
                self.add_epoch(time_var, size_vars,
                               mig_vars, dyn_vars, sel_vars)
            if n_pop < len(structure):
                if self.frac_split:
                    frac_var = FractionVariable(f"s{n_pop}")
                    size_split = copy.copy(size_vars[-1])
                    size_1 = Multiplication(frac_var, size_split)
                    size_2 = Multiplication(Subtraction(1, frac_var),
                                            size_split)
                    size_vars = [size_1, size_2]
                else:
                    name = size_vars[-1].name
                    size_vars = [PopulationSizeVariable(name + '_1'),
                                 PopulationSizeVariable(name + '_2')]
                self.add_split(n_pop - 1, size_vars)
        return self

    def get_structure(self):
        """
        Returns current structure of the model.
        """
        structure = [1]
        for event in self.events:
            if isinstance(event, Split):
                structure.append(0)
            elif isinstance(event, Epoch):
                structure[-1] += 1
            else:
                raise ValueError("Event is not Split or Epoch.")
        return structure

    def increase_structure(self, new_structure=None, X=None):
        """
        Increase structure of the model. Raises ValueError if structure is
        equal or greater than `final_structure`.

        :param new_structure: New structure for the model. Should be greater
                              by 1 in one element. E.g. structure is (1,2),
                              then new structure could be either (2,2) or
                              (1,3). If None random available element is
                              chosen to increase.
        :param X: list of values to transform as values of new model.

        :note: Function is specific for\
               :class:`gadma.models.StructureDemographicModel`.\
               It probably will not work for any other class.
        """
        cur_structure = self.get_structure()
        diff = np.array(self.final_structure) - np.array(cur_structure)
        if np.any(diff < 0):
            raise ValueError(f"Demographic model has its final structure "
                             f"({list(self.final_structure)}). It is not "
                             f"possible to increase it")
        # TODO check that new structure is not greater than final structure
        if new_structure is None:
            struct_index = np.random.choice(
                np.arange(len(cur_structure))[diff != 0])

            new_structure = copy.copy(cur_structure)
            new_structure[struct_index] += 1
        else:
            diff = np.array(new_structure) - np.array(cur_structure)
            if np.max(diff) < 1:
                raise ValueError(f"New structure ({new_structure}) should be "
                                 f"greater than the current one "
                                 f"({cur_structure})")
            if np.min(diff) > 1 or np.sum(diff) > 1:
                raise ValueError(f"New structure ({new_structure}) should "
                                 f"differ in one value and maximum by 1. "
                                 f"Received structure: {new_structure}")
            struct_index = np.where(diff == 1)[0]
            assert len(struct_index) == 1
            struct_index = struct_index[0]

        event_index = np.random.choice(np.arange(cur_structure[struct_index]))
        event_index += sum(cur_structure[:struct_index]) - 1 + struct_index

        old_model = copy.deepcopy(self)
        self.from_structure(new_structure)
        if X is None:
            return self, None

        # We consider that we have put new event (interval) before the chosen
        # event. Special case is when it is first interval - we put new event
        # after it.
        if event_index == -1:
            new_event = self.events[0]
            new_values = []
            for var in new_event.variables:
                if var not in new_event.get_vars_not_in_init_args():
                    continue
                # We put as time some random value
                if isinstance(var, TimeVariable):
                    new_values.append(new_event.time_arg.resample())
                # We put size of population as 1.0
                elif isinstance(var, PopulationSizeVariable):
                    new_values.append(1.0)
                # Dynamic as Sud
                elif isinstance(var, DynamicVariable):
                    new_values.append("Sud")
                elif isinstance(var, SelectionVariable):
                    new_values.append(0)
                else:
                    raise ValueError(f"Unknown type of variable: "
                                     f"{var.__class__}")
            new_X = []
            for x in X:
                new_X.append(copy.copy(new_values))
                new_X[-1].extend(x)
            return self, new_X

        # So we build dict with variables correspondence - they will be moved
        # forward by one event starting with chosen event.
        oldvar2newvar = {}
        for i, (old_event, new_event) in enumerate(zip(old_model.events,
                                                       self.events)):
            if i >= event_index:
                break
            for old_var, new_var in zip(old_event.variables,
                                        new_event.variables):
                oldvar2newvar[old_var] = new_var
        for old_event, new_event in zip(old_model.events[event_index:],
                                        self.events[event_index + 1:]):
            # Remove init_sizes variables from list of vriables.
            # Then both lists will have the same length and have special order
            if isinstance(old_event, Split):
                old_vars = [var for var in old_event.variables]
                new_vars = [var for var in new_event.variables]
            else:
                old_vars = old_event.get_vars_not_in_init_args()
                new_vars = new_event.get_vars_not_in_init_args()
            assert len(old_vars) == len(new_vars)
            # Now we cretae correspondence between those variables
            for old_var, new_var in zip(old_vars, new_vars):
                assert type(old_var) == type(new_var)  # addit. check for types
                oldvar2newvar[old_var] = new_var
    #    print(oldvar2newvar)
        new_X = []
        for x in X:
            # Our initial var2value
            var2value = old_model.var2value(x)
            # Now we get new var2value. We should be carefull as variables
            # in new_model have the same names but they are different from
            # those in model.
            varname2value = {var.name: var2value[var] for var in var2value}
            new_var2value = {var: varname2value[var.name]
                             for var in self.variables
                             if var.name in varname2value}
            for var in var2value:
                new_var2value[oldvar2newvar[var]] = var2value[var]
            event1 = self.events[event_index]  # our new event
            event2 = self.events[event_index + 1]  # base event
            # Time / 2
            new_var2value[event2.time_arg] /= 2
            new_var2value[event1.time_arg] = new_var2value[event2.time_arg]
            time_value = new_var2value[event1.time_arg]

            # Sizes
            for i, (size_in, size_out) in enumerate(zip(event1.init_size_args,
                                                        event1.size_args)):
                if event2.dyn_args is not None:
                    dyn_value = new_var2value[event2.dyn_args[i]]
                else:
                    dyn_value = 'Sud'
                if dyn_value != 'Sud':
                    func = DynamicVariable.get_func_from_value(dyn_value)
                    if event_index == 0:
                        y1 = 1.0
                    else:
                        # Init size could be with fraction.
                        if not isinstance(size_in, Variable):
                            vals = [new_var2value[var]
                                    for var in size_in.variables]
                            y1 = size_in.get_value(vals)
                        else:
                            y1 = new_var2value[size_in]
                    y2 = new_var2value[size_out]
                    # We have already divided it.
                    x_diff = 2 * time_value
                    size_func = func(y1, y2, x_diff)
                    new_var2value[size_out] = size_func(x_diff / 2)
            # Copy other variables
            for var1, var2 in zip(event1.variables, event2.variables):
                if var1 not in new_var2value:
                    new_var2value[var1] = new_var2value[var2]
            new_X.append([new_var2value[var] for var in self.variables])

        return self, new_X

    def transform_values_from_other_model(self, model, x):
        assert isinstance(model, StructureDemographicModel)
        assert np.all(self.get_structure() == model.get_structure())

#        print('other', model.variables)
#        print('self', self.variables)
#        print('x', x)
        other_var2value = model.var2value(x)
        other_varname2value = {var.name: other_var2value[var]
                               for var in other_var2value}
        var2value = {}
        # First fill with common variables
        for var in self.variables:
            if isinstance(var, MigrationVariable):
                if self.have_migs == model.have_migs:
                    if self.sym_migs == model.sym_migs:
                        var2value[var] = other_varname2value[var.name]
            elif isinstance(var, SelectionVariable):
                if self.have_sels == model.have_sels:
                    var2value[var] = other_varname2value[var.name]
            elif isinstance(var, DynamicVariable):
                if self.have_dyns == model.have_dyns:
                    var2value[var] = other_varname2value[var.name]
            elif isinstance(var, FractionVariable):
                if self.frac_split == model.frac_split:
                    var2value[var] = other_varname2value[var.name]
            elif var.name in other_varname2value:
                var2value[var] = other_varname2value[var.name]

        # Transform other values
        varname2value = {}
        for var in self.variables:
            if var.name in varname2value:
                var2value[var] = varname2value[var.name]
            if var in var2value:
                continue
            if isinstance(var, MigrationVariable):
                if not model.have_migs:
                    var2value[var] = 0
                elif self.sym_migs != model.sym_migs:
                    ij = var.name.split('_')[-1]
                    assert len(ij) == 2
                    sym_mig_name = var.name[:-2] + ij[::-1]
                    if self.sym_migs and not model.sym_migs:
                        var2value[var] = ((other_varname2value[var.name] +
                                           other_varname2value[sym_mig_name]) /
                                          2)
                    else:
                        var2value[var] = other_varname2value[var.name]
                        varname2value[sym_mig_name] = var2value[var]
            elif isinstance(var, SelectionVariable):
                assert self.have_sels and not model.have_sels
                var2value[var] = 0
            elif isinstance(var, DynamicVariable):
                assert self.have_dyns and not model.have_dyns
                var2value[var] = 'Sud'
            elif isinstance(var, FractionVariable):
                assert self.frac_split and not model.frac_split
                n_split = int(var.name[1:])
                ind_before_split = sum(self.get_structure()[:n_split]) - 1
                nu1_before_split_name = "nu%d%d" % (ind_before_split, n_split)
                nu1_after_split_name = "nu%d%d_1" % (ind_before_split,
                                                     n_split)
                if n_split == 1 and self.get_structure()[0] == 1:
                    size_before_split_name = 1.0
                    size_after_split_time = other_varname2value["nu_1"]
                else:
                    size_before_split_name = other_varname2value[
                        nu1_after_split_name]
                    size_after_split_time = other_varname2value[
                        nu1_before_split_name]
                fraction = size_after_split_time / size_before_split_name
                fraction = max(fraction, var.domain[0])
                fraction = min(fraction, var.domain[1])
                var2value[var] = fraction
            elif isinstance(var, PopulationSizeVariable):
                assert not self.frac_split and model.frac_split
                assert len(var.name.split("_")) > 1

                nu_before_split_name = var.name[:-2]  # remove last _1 or _2

                # if there was ancestral population split there is no variable
                if (nu_before_split_name == "nu" and
                        nu_before_split_name not in varname2value):
                    pop_size = 1.0
                    frac_name = "s1"
                else:
                    pop_size = other_varname2value[nu_before_split_name]
                    frac_name = "s" + nu_before_split_name[-1]
                fraction = other_varname2value[frac_name]

                if var.name[-1] == "1":
                    var2value[var] = fraction * pop_size
                elif var.name[-1] == "2":
                    var2value[var] = (1 - fraction) * pop_size
            else:
                raise ValueError("Some changes in demographic models are not "
                                 "allowed or implemented. Got new variable "
                                 f"({var}) that cannot be processed.")
        return [var2value[var] for var in self.variables]
